#       
#           AutoDock | Raccoon2
#
#       Copyright 2013, Stefano Forli
#          Molecular Graphics Lab
#  
#     The Scripps Research Institute 
#           _  
#          (,)  T  h e
#         _/
#        (.)    S  c r i p p s
#          \_
#          (,)  R  e s e a r c h
#         ./  
#        ( )    I  n s t i t u t e
#         '
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#


import CADD.Raccoon2
# Raccoon stuff
import CADD.Raccoon2.HelperFunctionsN3P as hf
import RaccoonBasics as rb
import RaccoonEvents
import RaccoonPmvCamera 


import os, Pmw, glob
from PmwOptionMenu import OptionMenu as OptionMenuFix

import Tkinter as tk
import tkMessageBox as tmb
import tkFileDialog as tfd
from PIL import Image, ImageTk
import sys

# mgl modules
from mglutil.events import Event, EventHandler
from mglutil.util.callback import CallbackFunction #as cb
from mglutil.gui.BasicWidgets.Tk.thumbwheel import ThumbWheel

class ConfigTab(rb.TabBase, rb.RaccoonDefaultWidget):
    
    def __init__(self, app, parent, debug=False):
        rb.TabBase.__init__(self, app, parent, debug = False)
        rb.RaccoonDefaultWidget.__init__(self, parent)
        self.resource = self.app.resource
        self.dockengine = self.app.dockengine
        self.app.eventManager.registerListener(RaccoonEvents.SetDockingEngine, self.handleEngine)
        self.app.eventManager.registerListener(RaccoonEvents.ReceptorListChange, self.setReceptorList)
        self.app.eventManager.registerListener(RaccoonEvents.ReceptorListChange, self.nuke)

        self.camera = None # embedded camera object
        self._loaded_struct = []

        self._boxevent = RaccoonEvents.SearchConfigChange()
        self._searchevent = None # XXX TODO?
        
        # switch for loading multiple receptors in the viewer
        self.recLoaderMultiMode = tk.BooleanVar(value=False)
        self._currRec = []#  [ { 'name':xxxx, 'obj':[ rigid, flex ], 'fname':xxxx }, ... ]

        self.initIcons()
        self.makeInterface()



    def initIcons(self):
        """ initialize the icons for the interface"""
        icon_path = CADD.Raccoon2.ICONPATH
        f = icon_path + os.sep + 'system.png'
        self._ICON_sys = ImageTk.PhotoImage(Image.open(f))

        f = icon_path + os.sep + 'center_mol.png'
        self._ICON_center_mol = ImageTk.PhotoImage(Image.open(f))

        f = icon_path + os.sep + 'center_all.png'
        self._ICON_center_all = ImageTk.PhotoImage(Image.open(f))

        f = icon_path + os.sep + 'center_box.png'
        self._ICON_center_box = ImageTk.PhotoImage(Image.open(f))

        f = icon_path + os.sep + 'water.png'
        self._ICON_water = ImageTk.PhotoImage(Image.open(f))

        f = icon_path + os.sep + 'floppy.png'
        self._ICON_floppy = ImageTk.PhotoImage(Image.open(f))

        f = icon_path + os.sep + 'open.png'
        self._ICON_open = ImageTk.PhotoImage(Image.open(f))

        f = icon_path + os.sep + 'default.gif'
        self._ICON_default = ImageTk.PhotoImage(Image.open(f))

        f = icon_path + os.sep + 'singleMol.png'
        self._ICON_single = ImageTk.PhotoImage(Image.open(f))

        f = icon_path + os.sep + 'multiMol.png'
        self._ICON_multi = ImageTk.PhotoImage(Image.open(f))


    def handleEngine(self, event=None):
        """ handle widget display for different engines """
        #print "DD_CONFIG> EVENT notified that docking engine is now [%s]" % event.dockingengine
        self.dockengine = event.dockingengine


    def openConfig(self, fname = None):
        """ import data from a vina config """
        ad_ext = ("*.dpf", "*.DPF")
        ag_ext = ("*.gpf", "*.GPF")
        vina_ext = ("*.conf", "*.CONF")
        anyfile = ("Any file...", ("*"))
        if fname == None:
            if self.dockengine == 'vina':
                t = 'Select Vina config file to import'
                ftypes = [("Vina config file", vina_ext), anyfile ]                
            elif self.dockingengine == 'ad':
                t = 'Select AutoDock/AutoGrid parameter files to import'
                ftypes = [ ("AutoDock parameter file", ad_ext),
                           ("AutoGrid parameter file", ag_ext),
                            anyfile ]                
            fname = tfd.askopenfilename(parent=self.frame, title=t, filetypes = ftypes)
            if not fname: return
        #print "GOT FNAME[%s]" % fname

        if not fname: return
        # parse vina config
        if self.dockengine == 'vina':
            accepted = self._loadvinaconf(fname)
        # parse ad configs
        elif self.dockingengine =='ad':
            name, ext = os.path.splitext(fname)
            if ext.lower() == 'dpf':
                accepted = self._loaddpf(fname)
            elif ext.lower() == 'gpf':
                accepted = self._loadgpf(fname)
        t = 'Docking parameters'
        if len(accepted):
            m = ('The following docking parameters '
                 'have been accepted:\n\n%s' ) % ("\n".join(accepted))
            i = 'info'
        else:
            m = ('No docking parametes have been imported!\n'
                 'Inspect the config file and try again')
            i = 'error'
        tmb.showinfo(parent=self.frame, title=t, message=m, icon=i)
        self.app.eventManager.dispatchEvent(self._boxevent)
        # XXX TODO
        # self.app.eventManager.dispatchEvent(self._searchevent)
            

    def _loadvinaconf(self, fname, quiet=False):
        """ actual parser of vina CONF file """
        accepted = []
        # load data
        config = self.app.engine.gridBoxFromConf(fname)
        if not config:
            if not quiet:
                t = 'Config file error'
                m = ('There was an error loading the config file.\n'
                     'Inspect it and try again...')
                i = 'error'
                tmb.showinfo(parent=self.frame, title=t, message=m, icon=i)
            return False
        # center 
        center = [ config[k] for k in 'center_x', 'center_y', 'center_z' ]
        if self.app.engine.setBoxCenter(center):
            for i in range(3):
                self._thumbw_array[i].set( center[i])
            accepted.append('center')
        # size
        size = [ config[k] for k in 'size_x', 'size_y', 'size_z' ]
        if self.app.engine.setBoxSize(size):
            for i in range(3):
                self._thumbw_array[i+3].set( size[i])
            accepted.append('size')
        # search parms
        for k in ['exhaustiveness', 'num_modes', 'energy_range']:
            if not config[k] == None:
                v = config[k]
                self.app.engine.vina_settings[k] = v
                self.vinaWidgets[k].setvalue(v)
                accepted.append(k)
        # dispatch config EVENT
        #event = RaccoonEvents.SearchConfigChange()
        self.app.eventManager.dispatchEvent(self._boxevent)
        return accepted

    def _loaddpf(self, fname):
        """ """
        # load data
        # set widgets
        pass


    def _loadgpf(self, fname):
        """ """
        # load data
        # set widgets
        pass




    
    def saveConfig(self):
        """ """
        pass

    def setWheelBox(self, wheelname, value):
        """ wheel callback that sets center/size values in
            the docking engine
        """
        cx, cy, cz = self.app.engine.box_center
        sx, sy, sz = self.app.engine.box_size
        if wheelname=='center_x':
            cx = value
        elif wheelname=='center_y':
            cy = value
        elif wheelname=='center_z':
            cz = value
        elif wheelname=='size_x':
            sx = value
        elif wheelname=='size_y':
            sy = value
        elif wheelname=='size_z':
            sz = value
        self.app.engine.box_center = [ cx, cy, cz ]
        self.app.engine.box_size = [ sx, sy, sz ]

        for i in self.app.engine.box_center :
            if i == None:
                return
        self.Box3D.Set(center = (cx, cy, cz))
        self.Box3D.Set(xside=sx, yside=sy , zside=sz)
        self.app.eventManager.dispatchEvent(self._boxevent)

        # if a full triplet is defined, sync the data with the engine config internals
        if not None in self.app.engine.box_center:
            self.app.engine.setBoxCenter( self.app.engine.box_center )
        if not None in self.app.engine.box_size:
            self.app.engine.setBoxSize( self.app.engine.box_size )

    def makeInterfaceLocal(self):
        """filler for now"""
        pass
    
    def makeInterfaceOpal(self):
        """ """
        pass

    def makeInterface(self):
        """ create interface for ssh """
        self.resetFrame()
        
        colors = { 'center_x' : '#ff3333',
                   'center_y' : 'green',
                   'center_z' : '#00aaff',
                   'size_x' : '#ff3333',
                   'size_y' : 'green',
                   'size_z' : '#0099ff',
                   }

        frame_set = {  'ring_bd' : 1, 'ring_highlightbackground' :'black', 'ring_borderwidth' : 2, 
                    'ring_highlightcolor' : 'black', 'ring_highlightthickness' : 1, 'ring_relief' : 'flat', 
                    'groupchildsite_bg':'white', 'groupchildsite_relief':'sunken','ring_bg':'white',
                        'tag_bd' : '1', 'tag_highlightbackground' :'black', 'tag_borderwidth':2, 
                        'tag_highlightcolor' : 'black', 'tag_highlightthickness': 1}

        frame_set = { 'ring_bd' : 1, 'ring_highlightbackground' :'black', 'ring_borderwidth' : 2, 
                    'ring_highlightcolor' : 'black', 'ring_highlightthickness' : 1, 'ring_relief' : 'flat'}

        frame_set = {}
        bset = { 'bg' : '#95bed5', 'width' : 22, 'height': 22, 'relief' : 'raised'}
        bset = { 'bg' : '#2e363b', 'width' : 22, 'height': 22, 'relief' : 'raised'}
        bset = { 'bg' : '#a6abae', 'width' : 22, 'height': 22, 'relief' : 'raised'}
        bset = { 'bg' : '#969b9d'  } # 'width' : 22, 'height': 22, 'relief' : 'raised'}
        bset = {}
        bset.update(self.BORDER)

        # left frame
        lframe = tk.Frame(self.frame)
        # button tollbar 
        tk.Frame(lframe, height=7).pack(expand=0, fill='x', padx=0, pady=0, side='top', anchor='n')
        # minispacer

        bframe = tk.Frame(lframe)
        b = tk.Button(bframe, text='Load...', command = self.openConfig, image= self._ICON_open, compound='left',
            height=14,**bset)
        b.pack(expand=1,fill='x', anchor='w', side='left',padx=1)

        b.pack(expand=1,fill='x', anchor='w', side='left',padx=1)
        bframe.pack(expand=0, fill='x', padx=0, pady=0, side='top', anchor='n')
        ########################## center wheels
        self._thumbw_array = []
        c_group = Pmw.Group(lframe, tag_text='Center', tag_font=self.FONTbold, **frame_set)
        for lab in ['center_x', 'center_y', 'center_z']:
            cb = CallbackFunction(self.setWheelBox, lab)
            tw = ThumbWheel(
                c_group.interior(), labCfg={'text':lab, 'side':'left','fg':colors[lab],
                'bg':'black', 'width':9 }, showLabel=1,
                width=90, height=14, type=float, value=0.0,
                callback=cb, continuous=True,
                oneTurn=5, wheelPad=0)
            tw.pack(side='top', pady=2,anchor='n')
            self._thumbw_array.append(tw)
        c_group.pack(side='top', anchor='n', expand=0,fill='x',ipadx=2,ipady=3, padx=1)

        ########################## size wheels
        s_group = Pmw.Group(lframe, tag_text='Size', tag_font=self.FONTbold, **frame_set)
                     
        for lab in ['size_x', 'size_y', 'size_z']:
            cb = CallbackFunction(self.setWheelBox, lab)
            tw = ThumbWheel(
                s_group.interior(), labCfg={'text':lab, 'side':'left','fg':colors[lab],
                'bg':'black', 'width':9 }, showLabel=1,
                width=90, height=14, type=float, value=0.0, min=0.0001,
                callback=cb, continuous=True,
                oneTurn=5, wheelPad=0)
            tw.pack(side='top', pady=2,anchor='n')
            self._thumbw_array.append(tw)
        s_group.pack(side='top', anchor='n', expand=0,fill='x',ipadx=2,ipady=3, padx=1,pady=2)

        ########################## search settings
        self.searchparmgroup = Pmw.Group(lframe, tag_text='Search parameters', tag_font=self.FONTbold, **frame_set)
        # autodock search parms
        self.buildSearchADPanel(target = self.searchparmgroup )
        # vina search parms
        self.buildSearchVinaPanel(target = self.searchparmgroup )

        self.searchparmgroup.pack(side='top', anchor='n', expand=0,fill='x',ipadx=2,ipady=3, padx=1,pady=5)

        ########################## Receptors
        self._receptors_group = Pmw.Group(lframe, tag_text='Receptor list', tag_font=self.FONTbold,collapsedsize=3, **frame_set)
        # important
        self.receptorListbox = Pmw.ScrolledListBox(self._receptors_group.interior(), listbox_highlightbackground = 'black',
            # selectioncommand=self.loadreceptor, 
            listbox_selectbackground='yellow')
        self.receptorListbox.pack(expand=1, fill='both',padx=3, pady=0)
        self.receptorListbox.component('listbox').bind('<ButtonRelease-1>', self.loadreceptor)
        self.receptorListbox.component('listbox').bind('<Button-3>', self._delreceptor)

        tb = tk.Frame(self._receptors_group.interior())
        # single-multi load buttons
        self.recLoaderMode_single = tk.Radiobutton(tb, text='Single', image = self._ICON_single, indicatoron=False, 
            variable=self.recLoaderMultiMode, value = False, compound='left', height=16, **bset)
        self.recLoaderMode_single.pack(anchor='n', side='left',pady=1,expand=1,fill='x',padx=1)
        self.recLoaderMode_multi = tk.Radiobutton(tb, text='Multi', image = self._ICON_multi, indicatoron=False,
            variable=self.recLoaderMultiMode, value = True, compound='left', height=16, **bset)
        self.recLoaderMode_multi.pack(anchor='n', side='left',pady=1,expand=1,fill='x',padx=1)
        

        tb.pack(expand=0, fill='x',anchor='s',side='bottom',padx=3)

        self._receptors_group.pack(side='bottom', anchor='n', expand=1, fill='both',ipadx=4,ipady=4, padx=1,pady=0)

        lframe.pack(side = 'left', anchor='n', expand='n', fill='y', padx=0, pady=0)

        ###### 3D Viewer
        rframe = tk.Frame(self.frame)
        spacer = tk.Frame(rframe, width=5) #, bg='red')
        spacer.pack(expand=0,fill='y',side='left',anchor='w')
        spacer.pack_propagate(0)
        vgroup = Pmw.Group(rframe, tag_text = '3D viewer', tag_font=self.FONTbold,groupchildsite_bg='black',  **frame_set)

        # TOOLBAR
        vtoolbar = tk.Frame(vgroup.interior())
        vtoolbar.pack(side='left', anchor='w', expand=0, fill='y')
        cb = CallbackFunction(self.centerView, None)
        tk.Button(vtoolbar, text='Center\nall', image = self._ICON_center_all, width=22, height=22, 
            command=cb, **bset).pack(anchor='n', side='top')
        cb = CallbackFunction(self.centerView, 'mol')
        tk.Button(vtoolbar, text='Center\nmol', image = self._ICON_center_mol, width=22, height=22,
            command=cb, **bset ).pack(anchor='n', side='top')
        cb = CallbackFunction(self.centerView, 'box')
        tk.Button(vtoolbar, text='Center\nbox', image = self._ICON_center_box, width=22, height=22,
            command=cb, **bset ).pack(anchor='n', side='top',pady=1)
        # 3d Viewer settings XXX TODO
        #tk.Button(vtoolbar, text='Settings', image = self._ICON_sys, width=22, height=22, **bset).pack(anchor='n', side='top')
        vgroup.pack(side='right', anchor='e', expand=1, fill='both', padx=0, pady=0)

        # 3D viewer  
        self.make3Dboxviewer(vgroup.interior())

        rframe.pack(side = 'right', anchor='n', expand=1, fill='both',padx=0, pady=0)
        
        if self.app.dockengine == 'vina':
            self.setSearchParmsVina()
        elif self.app.dockengine == 'autodock':
            self.setSearchParmsAD()
        self.frame.pack(expand=1, fill='both',anchor='n', side='top')

    def _delreceptor(self, event=None):
        """ """
        return
        # XXX TODO TO REMOVE RECEPTORS FROM HERE
        parent = self.receptorListbox 
        try:
            sel = parent.getvalue()
            #print "SEL", sel
            menu = tk.Menu(parent=parent, tearoff=False, takefocus=1)
            txt = 'Delete "%s"' % sel
            menu.add_command(txxt, state='normal', command = self.receptor)
        except:
            #print "nothing"
            pass


    def make3Dboxviewer(self, target):
        """ create the config camera and add the default box object"""
        # grid cam name 
        name = 'gridcam'
        if self.app.viewer == None:
            # XXX reinitialize fonts?
            self.app.viewer = RaccoonPmvCamera.EmbeddedMolViewer()
            self.app.molviewer = self.app.viewer.mv
            self.app.parent.bind_all('<F12>', self.app.viewer.togglepmv)
        # adding config camera
        if self.camera == None:
            self.camera = self.app.viewer.addCamera(target, name)
            self.camera_name = name
            from DejaVu.Box import Box
            from MolKit import Read
            self.Box3D = Box('Grid_box')
            self.Box3D.Set(xside=0.001, yside=0.001, zside=0.001)
            self.app.molviewer.GUI.VIEWER.AddObject(self.Box3D)        
            self.app.viewer.cameras[name]['structures'].append(self.Box3D)


    def centerView(self, target=None, event=None):
        """ center view in the 3D config viewer"""
        #root = self.app.viewer.rootObject
        showBox = False
        if target == 'box':
            print "Center box"
            item = self.Box3D
        elif target == 'mol':
            print "center mol"
            self.Box3D.visible = False
            item = None
            showBox = True
        else:
            print "CENTER ALL"
            item = None
        self.app.viewer.centerView(item)
        if showBox:
            self.Box3D.visible = True
            

    def getRecLoaderMode(self):
        """ check the value of variable self.recLoaderMultiMode 
            True  : keep adding receptors to the current
            False : delete previous receptors upon loading of the selected one
        """
        return self.recLoaderMultiMode.get()


    def loadreceptor(self, event=None):
        """ load the receptor in the viewer deleting previous mols
            if necessary
        """
        # TOP
        # TOP ######################
        self.receptorListbox.component('listbox').unbind('<ButtonRelease-1>')
        try:
            recname = self.receptorListbox.getvalue()[0]
        except:
            self.receptorListbox.component('listbox').bind('<ButtonRelease-1>', self.loadreceptor)
            return
        mode = 'single'
        if self.recLoaderMultiMode.get() == 1:
            mode = 'multi'
        cam = self.camera_name
        viewer = self.app.viewer
        #print "Loading rec", recname,
        #print "MODE", mode

        self.app.setBusy()

        if mode == 'single' and len(self._currRec)>1:
            # delete non-main 
            for i in range(len(self._currRec)-1):
                obj = self._currRec[i+1]['obj']
                for o in obj: viewer.deleteInCamera(o,cam)
                self._currRec.pop(i+1)

        recfiles = self.app.engine.getRecFiles(recname)
        recfname = recfiles[0]

        for i in range(len(self._currRec)):
            rec = self._currRec[i]
            if (rec['name'] == recname) and (rec['fname'] == recfname):
                if i == 0:
                    # already loaded and main
                    self.receptorListbox.component('listbox').bind('<ButtonRelease-1>', self.loadreceptor)
                    self.app.setReady()
                    return
                else:
                    # already loaded but not main
                    # put selected mol as first
                    self._currRec = [ self._currRec[i] ] + self._currRec[:i] + self._currRec[i+1:]
                    # FIXME apply STYLES (main, secondary)
                    self.app.setReady()
                    return
        # the current main is not the selected one
        if mode == 'single' and len(self._currRec) > 0:
            obj = self._currRec[0]['obj']
            for o in obj: viewer.deleteInCamera(o, cam)
            self._currRec = []
        newrec = { 'name' : recname, 'fname' : recfname }
        newrec['obj'] = [ viewer.loadInCamera(f, cam) for f in recfiles ]
        self._currRec = [newrec] + self._currRec
        self.receptorListbox.component('listbox').bind('<ButtonRelease-1>', self.loadreceptor)
        self.app.setReady()

    def nuke(self, event=None):
        """ remove all currently loaded receptors"""
        # we need to protect the box!
        #print "NUKE CALLED"
        for rec in self._currRec:
            #print "DELETING REC", rec
            for o in rec['obj']:
                self.app.viewer.deleteInCamera(o, self.camera_name)
        self._currRec = []

            
    def buildSearchVinaPanel(self, target):
        """ create the panel with vina search settings""" 
        try:
            # avoid double packaging
            self.searchParmVinaPanel
            return
        except:
            f = self.searchParmVinaPanel = tk.Frame(target.interior())
            target.configure(tag_text = 'Search parameters [ VINA ]')

        self.vinaWidgets = {}
        # exhaustiveness
        tk.Label(f, text='exhaustiveness', width =15,anchor='e').grid(row=1,column=1,sticky='we',padx=5)
        self.searchParmVina_exhaustiveness = Pmw.EntryField( f, validate = hf.validatePosNonNullInt,
            entry_justify='right', entry_width=8)
        self.searchParmVina_exhaustiveness.grid(row=1, column=2,sticky='we',columnspan=1)
        def d(event=None): self.setDefSearchParmVina('exhaustiveness')
        tk.Button(f, text='D', image=self._ICON_default, command=d).grid(row=1,column=3, sticky='e')
        self.vinaWidgets['exhaustiveness'] = self.searchParmVina_exhaustiveness


        # num modes
        tk.Label(f, text='num.modes', width =15,anchor='e').grid(row=2,column=1,sticky='we',padx=5)
        self.searchParmVina_nummodes = Pmw.EntryField( f, validate = hf.validatePosNonNullInt,
            entry_justify='right', entry_width=8)
        self.searchParmVina_nummodes.grid(row=2, column=2,sticky='we',columnspan=1)
        def d(event=None): self.setDefSearchParmVina('num.modes')
        tk.Button(f, text='D',image=self._ICON_default, command=d).grid(row=2,column=3, sticky='e')
        self.vinaWidgets['num_modes'] = self.searchParmVina_nummodes

        # energy range
        tk.Label(f, text='energy range', width =15,anchor='e').grid(row=3,column=1,sticky='we',padx=5)
        self.searchParmVina_energyrange = Pmw.EntryField( f, validate = hf.validateFloatPos, 
            entry_justify='right', entry_width=8)
        self.searchParmVina_energyrange.grid(row=3, column=2,sticky='we',columnspan=1)
        def d(event=None): self.setDefSearchParmVina('energy range')
        tk.Button(f, text='D', image=self._ICON_default,command=d).grid(row=3,column=3, sticky='e')
        self.vinaWidgets['energy_range'] = self.searchParmVina_energyrange

        # executed only the first time
        self.setDefSearchParmVina()


    def buildSearchADPanel(self, target):
        """ create the panel with autodock search settings""" 
        try:
            # avoid double packaging
            self.searchParmADPanel
            return
        except:
            f = self.searchParmADPanel = tk.Frame(target.interior())
            target.configure(tag_text = 'Search parameters [ AD ]')

        tk.Label(f, text='GA SETTINGS SHOULD BE HERE', width =15,anchor='w').grid(row=2,column=1,sticky='we')
        self.setDefSearchParmAD()

        
    def setDefSearchParmVina(self, request = None):
        """ """
        
        #widgets =  [ self.searchParmVina_exhaustiveness,
        #             self.searchParmVina_nummodes,
        #             self.searchParmVina_energyrange ]
        values = []
        wid = []
        if request in ['exhaustiveness', None]:
            values.append(8)
            wid.append(self.vinaWidgets['exhaustiveness'])
        if request in ['num_modes', None]:
            values.append(9)
            wid.append(self.vinaWidgets['num_modes'])
        if request in ['energy_range', None]:
            values.append(3)
            wid.append(self.vinaWidgets['energy_range'])
        for i in range(len(values)):
            wid[i].setentry(values[i])        

    def setDefSearchParmAD(self, value = None):
        #print "NO DEFAUILT SETTINGS FOR AD YET"
        return
        

    def setSearchParmsVina(self):
        """ pack and activate the search parms panel for vina """
        self.searchParmADPanel.pack_forget()
        self.searchParmVinaPanel.pack(expand=0,fill='x',anchor='n',padx=0, pady=0)


    def setSeachParmsAD(self):
        """ pack and activate the search parms panel for autodock """
        self.searchParmVinaPanel.pack_forget()
        self.searchParmADPanel.pack(expand=0,fill='x',anchor='n',padx=0, pady=0)


    def getSearchParmVina(self):
        """ """
        widgets =  [ self.searchParmVina_exhaustiveness, self.searchParmVina_nummodes,
                     self.searchParmVina_energyrange ]
        for w in widgets:
            if not w.valid():
                return False
        return  { 'exhaustiveness': self.searchParmVina_exhaustiveness.getvalue(),
                  'num.modes' : self.searchParmVina_nummodes.getvalue(),
                  'energyrange': self.searchParmVina_energyrange.getvalue(),
                }

    def getSearchParmAD(self):
        """ """ 
        # walk through ad parms and return settings
        pass


    def setReceptorList(self, event=None):
        """update the list of receptors shown in the picker for the config
            settings
        """
        self.receptorListbox.clear()
        self.receptorListbox.setlist(sorted(self.app.engine.RecBook.keys()) )





#########################################################################
#########################################################################
# $Header: /opt/cvs/CADD/FormsCADD.py,v 1.4.4.1 2015/08/26 21:18:11 sanner Exp $
# $Id: FormsCADD.py,v 1.4.4.1 2015/08/26 21:18:11 sanner Exp $

import Tkinter
import sys, re, os
import string
from Tkinter import *

from mglutil.util.misc import ensureFontCase
from Vision.Forms import Dialog, ChangeFonts, BugReport, VarDict

class ColorDialog(Dialog):

    def __init__(self):
        Dialog.__init__(self)

    def setDialogColor(self, bg, fg):
        self.bg = bg
        self.fg = fg
        self.textFrame.config(bg = self.bg, fg = self.fg )


class AboutWF(Dialog, ColorDialog):

    def __init__(self, name='', text='', bg="white", fg="black", avg="#000044"):
        self.panel = None
        self.name = name
        self.text = text
        self.bg = bg
        self.fg = fg
	self.avg = avg

        self.buildPanel()
        self.panel.protocol("WM_DELETE_WINDOW", self.Ok)
        self.textFrame.configure(state='disabled')

    def resetTags(self):
        self.textFrame.tag_configure('email', font=(ensureFontCase('times'), 10, 'bold'),
                                     foreground=self.avg)
        self.textFrame.tag_configure('http', font=(ensureFontCase('times'), 12, 'bold'),
                                     foreground=self.avg)
        self.textFrame.tag_configure('normal10b',
                                     font=(ensureFontCase('helvetica'), 10, 'bold'),
                                     justify="right",
                                     foreground=self.avg)
        self.textFrame.tag_configure('mediumb',
                                     font=(ensureFontCase('helvetica'), 14, 'bold'),
                                     foreground=self.avg)
        self.textFrame.tag_configure('title',
                                     font=(ensureFontCase('helvetica'), 14, 'bold'),
                                     foreground=self.avg, justify="center")

    def buildPanel(self):
        Dialog.buildPanel(self) 
        self.scroll.forget()

        self.panel.title(self.name + ' workflows')
        self.textFrame.configure()
        self.textFrame.insert('end', self.text, 'normal12')
        self.textFrame.config(bg = self.bg, fg = self.fg, height=15, width=100)
	self.resetTags()


class AboutDialogCADD(AboutWF):

    def __init__(self, bg, fg):
        AboutWF.__init__(self, bg = bg, fg = fg)

    def buildPanel(self):
        AboutWF.buildPanel(self) 

        # remove scrollbar
        self.scroll.forget()

        # add own stuff
        self.panel.title('About CADD')

        self.textFrame.configure()

        self.textFrame.insert('end', 'CADD  Pipeline\n\n', 'title')
        self.textFrame.insert('end', 
                              'The Computer Aided Drug Discovery (CADD) Pipeline is a workflow environment\n'+\
			      'designed to support  Molecular  Dyanmics  Simulations and  Virtual  Screening\n'+\
			      'experiments for in silico drug discovery.\n\n',
                              'normal12')
        self.textFrame.insert('end',
                              'CADD pipeline is released as a set of Vision networks packaged for specific processes.\n\n', 
                               'normal12')
        self.textFrame.insert('end', 'Home page:\t', 'normal12')
        self.textFrame.insert('end', 
                              'http://www.nbcr.net/wiki/index.php?title=CADD_Pipeline\n\n',
                              'http')

class AcknowlDialogCADD(AboutWF):
    def __init__(self, bg, fg):
        AboutWF.__init__(self, bg = bg, fg = fg)

    def buildPanel(self):
        AboutWF.buildPanel(self) 

        # add own stuff
        self.panel.title('Acknowledgements')

        self.textFrame.insert('end', 'CADD Acknowledgements\n', 'title')
        self.textFrame.insert('end',
                              '\n\tTSRI\t\t\t\tUCSD\n\n', 'http')
        self.textFrame.insert('end',
                              'Sargis Dallakyan\t\t\t\tRommie Amaro\n'+\
                              'Stefano Forli\t\t\t\tLuca Clementi \n'+\
                              'Ruth Huey\t\t\t\tWilfred Li \n'+\
                              'Michel Sanner\t\t\t\tAndy Mccammon \n'+\
                              'Art Olson\t\t\t\tJane Ren (alumni) \n'+\
                              'Guillaume Vareille (alumni)\t\t\t\tNadyaWilliams', 'normal10')
        

class RefDialogCADD(AboutWF):

    def __init__(self, bg, fg):
        AboutWF.__init__(self, bg = bg, fg = fg)

    def buildPanel(self):
        AboutWF.buildPanel(self) 

        # remove scrollbar
        self.scroll.forget()

        # add own stuff
        self.panel.title('References')

        self.textFrame.insert('end', 'CADD References\n\n', 'title')

        self.textFrame.insert('end',
            'Opal web services for biomedical applications.\n','normal12') 
        self.textFrame.insert('end',
            'Ren, J., Williams N., Clementi L., Krishnan S. and Li W. ','normal10') 
        self.textFrame.insert('end',
            'Nucleic Acids Res 38: W724-31, 2010.\n\n','normal10')
        self.textFrame.insert('end',
            'Emerging methods for ensemble-based virtual screening.\n', 'normal12')
        self.textFrame.insert('end',
                              'Amaro, R. and Li W. ', 'normal10')
        self.textFrame.insert('end',
        'Curr Top Med Chem 10: 3-13, 2010.\n\n','normal10')


class ChangeFontsCADD(ChangeFonts):

    def __init__(self, editor=None):
	    ChangeFonts.__init__(self, editor=editor)

    def Apply_cb(self, event=None):
        if self.editor is not None:
            cfg = self.get()
            self.editor.setFont(cfg[0], cfg[1])
            from CADD.UserLibBuildCADD import saveFonts4CADDFile
            saveFonts4CADDFile(self.editor.font)
        

class BugReportCADD(AboutWF):
    def __init__(self, bg, fg):
        AboutWF.__init__(self, bg = bg, fg = fg)

    def buildPanel(self):
        AboutWF.buildPanel(self) 

        # remove scrollbar
        self.scroll.forget()

        # add own stuff
        self.panel.title('References')

        self.textFrame.insert('end', 'CADD Bug Report\n\n', 'title')

        self.textFrame.insert('end',
            'Please submit your bug report to admin@nbcr.net \n','http') 


class BugReportCADD2(BugReport):

    def __init__(self, title="Bug Report", message="", bg=None,
                 fg=None, font=(ensureFontCase('helvetica'), 10, 'bold'),):
        BugReport.__init__(self, title, message, bg, fg, font,)


    def show_upload_page(self):
        self.get_description()
         
        sumcont = VarDict['shortdesc_text']
        desccont = VarDict['desc_text']

        if len(sumcont)<=2 or len(desccont)<=2:
            import tkMessageBox
            ok = tkMessageBox.askokcancel("Input","Please enter summary and description")
            return
        from mglutil.TestUtil import BugReport
        if VarDict.has_key('attachfile'):
            upfile = VarDict['attachfile']
        else:
            upfile=[]
        
        print "VarDict 2 :",  VarDict

        BR = BugReport.BugReportCommand("CADD")
        if self.validateEmail(VarDict['email_recipient']):
            idnum = BR.showuploadpage_cb(sumcont,desccont,upfile,email_ent=VarDict['email_recipient'])
        else:
            idnum = BR.showuploadpage_cb(sumcont,desccont,upfile,email_ent="")
        
        self.Ok() 
        
        # Tk message Box
        root = Tk()
        t  = Text(root)
        t.pack()


        def openHLink(event):
            start, end = t.tag_prevrange("hlink",
                               t.index("@%s,%s" % (event.x, event.y)))
            webbrowser.open_new('%s' %t.get(start, end))
            #print "Going to %s..." % t.get(start, end)

        t.tag_configure("hlink", foreground='blue', underline=1)
        t.tag_bind("hlink", "<Control-Button-1>", openHLink)
        t.insert(END, "BugReport has been Submiited Successfully\n")
        t.insert(END, "BugId is %s" %idnum)
        t.insert(END,"\nYou can visit Bug at\n")
        t.insert(END,"http://mgldev.scripps.edu/bugs/show_bug.cgi?id=%i" %int(idnum),"hlink")
        t.insert(END,"\nControl-click on the link to visit this page\n")
        t.insert(END,"\n")




#       
#           AutoDock | Raccoon2
#
#       Copyright 2013, Stefano Forli
#          Molecular Graphics Lab
#  
#     The Scripps Research Institute 
#           _  
#          (,)  T  h e
#         _/
#        (.)    S  c r i p p s
#          \_
#          (,)  R  e s e a r c h
#         ./  
#        ( )    I  n s t i t u t e
#         '
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#!/usr/bin/env python


from mglutil.math.rmsd import RMSDCalculator
from numpy import array, zeros, sqrt  
from bhtree import * # bhtreelib
from CADD.Raccoon2.piStackingAndRingDetection import *  #(pi-stacking interatcion)
import time
import os, sys
from CADD.Raccoon2.HelperFunctionsN3P import getCoords, pmvAtomStrip, getLines, quickdist, dist, percent, getAtype # TODO
from CADD.Raccoon2.WaterProcessing import processHydroDocking

class VSDockingResultsGenerator:
    """VS result class for extracting the results (LE,LC) from a series of DLG files
       and calculate properties (L.eff, cluster statistics, receptor contacts)
       - flexible residue interactions are calculated
       - simple planar ring detection on ligand (and target co-factors) is implemented for pi-stackings
       - Pi-stackings (parallel, t-stack) interactions are detected

       if auto is True, the list of dlgs will be used immediately to run recluster results,
       otherwise methods recluster() and calcInteractions() must be used.

       Receptor structure can be specified as filename or as {text|array|atype} dictionary.

       if a filename is specified, the file is processed by self.getCoords() and the self.recname is set.

       if a pre-processed receptor dictionary is provided, self.recname must be specified explicitly.

       receptor = { 'text' : atoms, 'coord' : array( coord, 'f'), 'atype': atype }

       A simple usage could be this:

        result = VSDockingResultsGenerator(dlgList = dlg_list, rmsTol = 2.0, receptor = receptor_file, recname = None, auto = True)
        print result
        pdbqt = result.generatePDBQTplus()

    # print "ITERTOOLZ!" <- improve speed?
    # http://docs.python.org/library/itertools.html
    # http://docs.python.org/release/2.5.2/tut/node7.html
    # http://www.doughellmann.com/PyMOTW/itertools/

    """
    #
    # TODO Note on the self.receptor
    # initialize the object with self.recPiVectors:
    # i.e. result = VSDockingResultsGenerator(dlgList = list)
    #      result.recPiVectors(vectorsCached)
    # to save time from re-calculating every time the PiVectors
    # TODO ...or not necessary? Pi vectors should be attached to the receptor parsed
    #          coords?  {text : ...., pi_vectors : [ ] }

    def __init__(self, input_files = None,           # list of DLG[AD]/*_out.pdbqt[VINA] contaning poses
                       #rmsTol = None,          # clustering RMSD tolerance
                       recname = None,         # receptor name (if not provided it will be taken from 'self.receptor'
                       receptor = None,        # receptor coordinates: PDBQT or getCoord() output (CACHE)
                       doInteractions = True,  # calculate ligand-target interactions
                       debug = False,          # debug mode
                       hbtol = 0.0,            # hb distance extra tolerance
                       water_map = None,       # water map used for rescoring W atoms
                       fullentropy=0,          # include entropy penalty for conserved waters # TODO REMOVE THIS?
                       ):
        
        # default values
        self.DEBUG = debug

        if self.DEBUG:
            #print "VSDockingResultsGenerator> init:", input_files, recname , receptor, doInteractions, debug, hbtol, water_map
            if not hbtol == 0 :
                print "DEBUG: extra hbtolerance:", hbtol
        #print "VSDockingResultsGenerator> init:", input_files, recname , receptor, doInteractions, debug, hbtol, water_map
        
        HB_CUTOFF=3.21 + hbtol # X-Y hydrogen-bond distance cutoff -X-H...Y- # from AutoGrid 4.2 values
        VDW_SMOOTHING = .5 # vdw tolerance (account for AutoGrid map smoothing)
        # TODO TODO TODO TODO TODO
        # TODO these values must be adapted *differently* for VINA AND AUTODOCK? TODO TODO TODO 

        # /default values

        self.VDW_TOL = VDW_SMOOTHING
        self.HB_CUTOFF = HB_CUTOFF**2 # save time

        # from AD4_parameter.dat Rii/2 values
        self.vdw_radii = { 'H': 1.00, 'HD': 1.00, 'HS': 1.00, 'C': 2.00,
            'A': 2.00, 'N': 1.75, 'NA': 1.75, 'NS': 1.75, 'OA': 1.60,
            'OS': 1.60, 'F': 1.54, 'Mg': 0.65, 'MG': 0.65, 'P': 2.10,
            'SA': 2.00, 'S': 2.00, 'Cl': 2.04, 'CL': 2.04, 'Ca': 0.99,
            'CA': 0.99, 'Mn': 0.65, 'MN': 0.65, 'Fe': 0.65, 'FE': 0.65,
            'Zn': 0.74, 'ZN': 0.74, 'Br': 2.165, 'BR':2.165, 'I':2.36,
            'Z' : 2.00, 'G' : 2.00, 'GA': 2.00, 'J' :2.00, 'Q' :2.00,
            'W' : 3.0,  # TODO evaluate this value...
            'X': 2 } # default vdW for unknown atom

        # list of atoms that are ignored when calculating RMSD
        self.ignore_at = [ 'HD', 'H', 'W' ]
        #self.ignore_at = [ 'HD', 'H' ]
        #print " \n\n***\n\n Change the management of W atom types! contacts and so on...\notherwise they will not be correctly removed\n\n***\n\n"
        self.water_types = [ 'W' ]
        self.water_map = water_map
        self.isHydrated = False
        self.fullentropy = fullentropy
        self.recname = recname
        self.receptor = receptor
        self.doInteractions = doInteractions
        self.hbtol = hbtol

        # HB?
        #        self.hb_radii = {
        #'N'      3.50  0.160  22.4493  -0.00162  0.0  0.0  0  -1  -1  1	# Non H-bonding Nitrogen
        #'NA'     3.50  0.160  22.4493  -0.00162  1.9  5.0  4  -1  -1  1	# Acceptor 1 H-bond Nitrogen
        #'NS'     3.50  0.160  22.4493  -0.00162  1.9  5.0  3  -1  -1  1	# Acceptor S Spherical Nitrogen
        #'OA'     3.20  0.200  17.1573  -0.00251  1.9  5.0  5  -1  -1  2	# Acceptor 2 H-bonds Oxygen
        #'OS'     3.20  0.200  17.1573  -0.00251  1.9  5.0  3  -1  -1  2	# Acceptor S Spherical Oxygen
        #'SA'     4.00  0.200  33.5103  -0.00214  2.5  1.0  5  -1  -1  6	# Acceptor 2 H-bonds Sulphur
        #'S'      4.00  0.200  33.5103  -0.00214  0.0  0.0  0  -1  -1  6	# Non H-bonding Sulphur
        #}

        self.metals = [ 'Mg', 'MG', 'Mn', 'MN', 'Fe', 'FE', 'Zn', 'ZN', 'CA', 'Ca'] # metals
        self.met_coord = [ 'OA', 'OS', 'SA', 'S', 'N', 'NA', 'NS' ]                 # metal-binders

        if input_files:
            self.setLigands(input_files = input_files)

        if receptor:
            self.setReceptor(receptor=receptor, recname=recname)

        
    def setLigands(self, input_files=[]): 
        """ set the input files to be processed
        """
        self.input_files = input_files
        self.poses = []
        self.problematic = []
        self.ligName = None
        self.atomTypes = []
        self.results = []
        self.flexres = False
        self.interactions = False
        self.histogram = []
            

    def setReceptor(self, receptor, recname=None):
        # pre-process receptor if filename
        if isinstance(receptor, str) or isinstance(receptor, unicode):
            if recname == None:
                r = os.path.basename(receptor)
                recname = os.path.splitext(r)[0]
            try:
                receptor = getCoords(getLines(receptor))
            except:
                print "Problem in reading the receptor [%s]: (%s) " % (receptor, sys.exc_info()[1])
                return False

        self.receptor = receptor
        self.recname = recname

        # initialize the bh_tree for the 
        # interactions calculated later
        try:
            freeBHtree(self.receptor_bht)
        except:
            pass
        self.rec_bht           = bhtreelib.BHtree( self.receptor['coord'], None, 10)
        self.rec_bht_indices   = zeros( (len(self.receptor['coord']),) ).astype('i')
        self.rec_bht_distances = zeros( (len(self.receptor['coord']),) ).astype('f') 


    def getPath(self):
        """ return the path where the input data is stored
            by reading it from the first ligand file
        """
        if not self.input_files:
            return
        if isinstance(self.input_files,list):
            f = self.input_files[0]
        elif isinstance(self.input_files, str) or isinstance(self.input_files, unicode):
            f = self.input_files
        path = os.path.dirname(f)
        return path

    
    def findHbAccepDon(self, atom_list):
        """identifies HB donors and acceptors in a list of PDBQT atoms
           returns : acceptors[] and donors[] lists
        """
        H_COV_BOND = 1.19 # value adapted for some distorted structures (ZINC21002974)
        H_COV_BOND  = H_COV_BOND ** 2  
        acceptor_types = ['OA', 'NA', 'SA']
        donor_types = ['N', 'O', 'OA', 'NA', 'SA']
        acceptors = []
        donors = []
        h = []
        dcandidate = []
        for l in atom_list:
            if l.startswith("ATOM") or l.startswith("HETATM"):
                l = l.strip()
                atype=l.split()[-1]
                if atype in acceptor_types:
                    if not l in acceptors:
                        acceptors.append(l)
                if atype in donor_types:
                    if not l in dcandidate:
                        dcandidate.append(l)
                if atype == 'HD':
                    if not l in h:
                        h.append(l)
        for a in dcandidate:
            for x in h:
                if dist(a, x) <= H_COV_BOND:
                    donors.append(a)
                    break

        return acceptors, donors 

    def calcInteractions(self):
        if not self.results:
            if self.DEBUG: print "Warning: no results."
            return False
        self.getContactAtoms()
        self.getHbInteractions()

        # XXX think about the optimal place for water processing
        # XXX
        # water processing happens after 
        # close contacts and HB acc/don are known...
        if self.isHydrated: 
            if self.water_map:
                self.processWaters(poses = self.results)
            else:
                print " # WARNING: inaccurate energy! [%s] hydrated ligand found, but no water map specified" %  self.ligName

        self.getPiStackInteraction()
        self.getMetalCoordInteraction()
        self.interactions = True

    # Phase 2 functions
    def getContactAtoms(self):
        """ populate self.result[]['vdw_contacts'] list 
            saves receptor metals for potential usage later (metal-coordination)
        """
        cutoff    = 5. # initial cutoff for bhtree
        #print len(self.receptor['coord'])


        # XXX THIS PART COULD BE MOVED IN THE INIT/set receptor part
        #bht       = bhtreelib.BHtree( self.receptor['coord'], None, 10)
        #indices   = zeros( (len(self.receptor['coord']),) ).astype('i')
        #distances = zeros( (len(self.receptor['coord']),) ).astype('f')
        for r in xrange(len(self.results)):
            self.results[r]['metal'] = []
            # A. find ligand atoms close contacts:
            for i in xrange( len(self.results[r]["coord"])):
                try:
                    l_vdw = self.vdw_radii[ self.atomTypes[i] ] + self.VDW_TOL
                except:
                    l_vdw = self.vdw_radii['X'] + self.VDW_TOL
                    print "Warning! Unrecognized ligand atom type!", self.atomTypes[i]
                #print self.atomTypes[i], l_vdw,

                l_coord = self.results[r]["coord"][i]
                nb = self.rec_bht.closePointsDist( tuple(l_coord), 
                        cutoff, self.rec_bht_indices, self.rec_bht_distances)
                # 1. rigid receptor atoms
                for j in xrange(nb):
                    rec_index = self.rec_bht_indices[j]
                    rec_atom = self.receptor['text'][rec_index]
                    if not rec_atom in self.results[r]['vdw_contacts']: # slight speed up?
                        d = self.rec_bht_distances[j]
                        try:
                            r_vdw = self.vdw_radii[self.receptor['atype'][rec_index]]
                        except:
                            r_vdw = self.vdw_radii['X']
                            print "Warning! Unrecognized receptor atom type!"
                        if d <= ( r_vdw + l_vdw ): 
                            # vdW contact
                            self.results[r]['vdw_contacts'].append(rec_atom)   
                            # metal detector
                            if self.receptor['atype'][rec_index] in self.metals:
                                self.results[r]['metal'].append(rec_atom)      
                del nb

                # 2. flexible residues atoms
                if self.flexres: 
                    if self.DEBUG: print "get_ContactAtoms> calc_distances for FlexRes too!"
                    for fres_idx in range(len(self.results[r]['flex_res'])):
                        r_coord = self.results[r]['flex_res']['coord'][fres_idx]
                        r_vdw = self.vdw_radii[self.results[r]['flex_res']['atype'][fres_idx]]
                        r_atom = self.results[r]['flex_res']['text'][fres_idx]
                        if not r_atom in self.results[r]['vdw_contacts']:
                            if quickdist(l_coord, r_coord) <= ( r_vdw + l_vdw ):
                                # vdW contact (flex res)
                                self.results[r]['vdw_contacts'].append(r_atom) 
                                # metal detector (flex res; it shouldn't happen...?)
                                if self.receptor['atype'][rec_index] in self.metals:
                                    self.results[r]['metal'].append(rec_atom)  
                                if self.DEBUG:
                                    print "contact with flex_res atom:", r_atom
                                    print quickdist(l_coord, r_coord), r_vdw , l_vdw , r_vdw+l_vdw
            # B. find water close contacts
            if self.isHydrated:
                if self.DEBUG: print "getContactAtoms> calc.distances for W atoms too"
                self.results[r]['water_bridge_contacts'] = []
                for w in self.results[r]['water_bridge']:
                    contacts = []
                    w_coord = atomCoord(w)
                    w_vdw = 2.0 + self.VDW_TOL # XXX water radius
                    nb = self.rec_bht.closePointsDist(tuple(w_coord), cutoff, 
                            self.rec_bht_indices, self.rec_bht_distances)
                    for j in xrange(nb):
                        rec_index = self.rec_bht_indices[j]
                        rec_atom = self.receptor['text'][rec_index]
                        if not rec_atom in contacts:
                            d = self.rec_bht_distances[j]
                            try:
                                r_vdw = self.vdw_radii[self.receptor['atype'][rec_index]]
                            except:
                                r_vdw = self.vdw_radii['X']
                                print "Warning! Unrecognized receptor atom type!"
                            if d <= ( r_vdw + w_vdw ): 
                                contacts.append(rec_atom)   
                    del nb
                    self.results[r]['water_bridge_contacts'].append(contacts)

            self.results[r]['rec_hb_candidates'] = {}
            acc, don = self.findHbAccepDon(self.results[r]['vdw_contacts'])
            if self.DEBUG:
                writeList("rec_acceptors.pdb", acc)
                writeList("rec_donors.pdb", don)
            self.results[r]['rec_hb_candidates']['acc'] = acc
            self.results[r]['rec_hb_candidates']['don'] = don
        # XXX this part should move to the initialization and be called every time 
        # a new receptor is called...
        # freeBHtree(bht)


    def getHbInteractions(self):
        """
        input   : pdb lines (ligand, receptor)
        output  : list of acceptor/donor pairs (lig:rec) <= PDBQT+ format
        notes   : distance only is used to characterize HB (consistant with AG maps)
                  flex res atoms are handled in a transparent manner being added to:
                     self.results[r]['rec_hb_candidates']['acc'] = acc
                     self.results[r]['rec_hb_candidates']['don'] = don.
        """
        for i in range(len(self.results)):
            # use the close contact atoms to calculate the HB (faster)
            hb_acc = []
            hb_don = []
            # HB acceptors
            for a in self.results[i]['hba_atoms']: # ligand hb acceptor atoms
                ltype = a.rsplit()[-1] 
                stol=0.
                if ltype=='SA': stol += 3.5 # TODO CHECK THIS VALUE!
                    #print "A>",a, "[%2.2f]" % stol
                for r in self.results[i]['rec_hb_candidates']['don']:
                    #rtype = r.rsplit()[-1] # TODO check this value
                    rtype = getAtype(r)
                    if rtype=='SA': stol += 3.5
                    if dist(a,r,sq=False)-stol <= self.HB_CUTOFF:
                        #print  dist(a,r)-stol, self.HB_CUTOFF
                        #print  sqrt( dist(a,r)-stol), sqrt(self.HB_CUTOFF)
                        if not [a,r] in hb_acc:
                            hb_acc.append([a,r])
            # HB donors
            for a in self.results[i]['hbd_atoms']: # ligand hb donor atoms
                if a.rsplit()[-1]=='SA': stol=2
                else: stol=0
                for r in self.results[i]['rec_hb_candidates']['acc']:
                    if dist(a,r)-stol <= self.HB_CUTOFF:
                        if not [a,r] in hb_don:
                            hb_don.append([a,r])
            self.results[i]['hb'] = { 'acceptors' : hb_acc, 'donors': hb_don}
            # TODO WATERS GO HERE?
            # Water bridges
            if self.isHydrated:
                pass


    def getPiStackInteraction(self):
        # TODO process the receptor in advance, cache the vectors!
        receptor = self.receptor['text']
        for i in range(len(self.results)):
            if self.flexres:
                #pstack, tstack = findLigRecPiStack(self.results[i]['text'], receptor+self.results[i]['flex_res']['text'])
                # TODO fix the flex_res bug! This godedoesn'twork
                true_lig_atoms = list(set(self.results[i]['text']) - set(self.results[i]['flex_res']['text']))
                # DEBUG XXX
                #writeList(str(i)+"_flex.pdb", self.results[i]['flex_res']['text'], addNewLine=True)
                #writeList(str(i)+"_lig.pdb", true_lig_atoms)
                pstack, tstack = findLigRecPiStack(true_lig_atoms[i], receptor+self.results[i]['flex_res']['text'])
            else:
                pstack, tstack = findLigRecPiStack(self.results[i]['text'], receptor)
            self.results[i]['pi'] = { 'p-stack' : pstack , 't-stack': tstack}

    def getMetalCoordInteraction(self):
        # it could be incorporated with getHbInteractions leading
        # to getPolarInteractions, but in this way it's easier to 
        # maintain... probably.
        metal_coordination = []
        for i in range(len(self.results)):
            for a in self.results[i]['hba_atoms']: # ligand hb acceptor atoms
                for r in self.results[i]['metal']: # receptor metals
                    if dist(a,r) <= self.HB_CUTOFF:
                        if not [a,r] in metal_coordination:
                            metal_coordination.append([a,r])
            self.results[i]['metal_coord'] = metal_coordination

    def calcReferenceRms(self, reference, keepH=False, debug=False):
        ref = getCoords(getLines(reference), include_hydrogens = keepH)
        if not len(ref['coord']) == len(self.results[0]['coord']):
            print "[atoms mismatch] Warning! The reference stricture doesn't match the docked ligand!"
            print "Reference [ %d ] | Docked ligand [ %d ]" % (  len(ref['coord']), len(self.results[0]['coord']) )
            return 
        rmsdcalc = RMSDCalculator(ref['coord'])
        rmsd = [ rmsdcalc.computeRMSD(self.results[0]['coord']) ]
        if len(self.results) > 1:
            rmsd.append(rmsdcalc.computeRMSD(self.results[1]['coord']))
        if debug:
            dist_pool = []
            for px in self.poses:
                p = px['coord']
                d = []
                for i in range(len(p)):
                    x = quickdist(p[i], ref['coord'][i], sq=True)
                    d.append(x)
                    d.sort()
                dist_pool.append(d)
                #print "================="
                #print d
            return rmsd, dist_pool,self.poses
        else:
            return rmsd

    def getInteractSummary(self, pose=0, recOnly=False):
        """ collect and format the interaction information for the pose
            if recOnly, the summary will inlcude only one entry
            for receptor residues involved in interactions
        """
        interactions = {}
        pose = self.results[pose]
        if recOnly:
            if pose['hb']['acceptors']:
                interactions['hba'] = [ pmvAtomStrip(p[0]) for p in pose['hb']['acceptors'] ]
            if pose['hb']['donors']:
                interactions['hbd'] = [ pmvAtomStrip(p[0]) for p in pose['hb']['donors'] ]
            if pose['vdw_contacts']:
                interactions['vdw'] = [ pmvAtomStrip(p) for p in pose['vdw_contacts'] ]
            if pose['metal_coord']:
                interactions['metal'] = [ pmvAtomStrip(p[0]) for p in pose['metal_coord'] ]
            if pose['pi']['p-stack']:
                interactions['ppi'] = [ p[0] for p in pose['pi']['p-stack'] ]
            if pose['pi']['t-stack']:
                interactions['tpi'] = [ p[0] for p in pose['pi']['t-stack'] ]
        else:
            if pose['hb']['acceptors']:
                interactions['hba'] = [ (pmvAtomStrip(p[0]), pmvAtomStrip(p[1])) for p in pose['hb']['acceptors'] ]
            if pose['hb']['donors']:
                interactions['hbd'] = [ (pmvAtomStrip(p[0]), pmvAtomStrip(p[1])) for p in pose['hb']['donors'] ]
            if pose['vdw_contacts']:
                interactions['vdw'] = [ pmvAtomStrip(p) for p in pose['vdw_contacts'] ]
            if pose['metal_coord']:
                interactions['metal'] = [ (pmvAtomStrip(p[0]), pmvAtomStrip(p[1])) for p in pose['metal_coord'] ]
            if pose['pi']['p-stack']:
                for n, ppi in enumerate(pose['pi']['p-stack']):
                    res, coordL, coordR = ppi
                    pose['pi']['p-stack'][n] = [ res, list(coordL), list(coordR) ]
                interactions['ppi'] = [ pose['pi']['p-stack'] ]
                #print "\n\n======================\n%s\n===================\n\n" % interactions['ppi']
            if pose['pi']['t-stack']:
                for n, ppi in enumerate(pose['pi']['t-stack']):
                    res, coordL, coordR = ppi
                    pose['pi']['t-stack'][n] = [ res, list(coordL), list(coordR) ]
                interactions['tpi'] = [ pose['pi']['t-stack'] ]

        return interactions



################################################## 
################################################## AutoDock class 
################################################## 

class AutoDockVsResult(VSDockingResultsGenerator):

    def __init__(self, input_files = None, rmsTol = None, 
                recname = None, receptor = None, auto = True, 
                doInteractions = True, hbtol = 0.0, 
                ignoreWaters = False, # XXX
                water_map = None,
                fullentropy=0,
                debug = False):
        VSDockingResultsGenerator.__init__(self,
            input_files = input_files,          # list of DLG[AD] from wich poses will be extracted
            recname = recname,                  # receptor name (if not provided it will be taken from 'self.receptor'
            receptor = receptor,                # receptor coordinates: PDBQT or getCoord() output (CACHE)
            doInteractions = doInteractions,    # calculate ligand-target interactions
            debug = debug,                      # debug mode
            hbtol = hbtol,
            water_map = water_map,
            fullentropy=fullentropy,
            )
        # defaults
        DEFAULT_RMSTOL = 2.0

        
        # AutoDoc specific properties
        self.tag = ["LE", "LC"]
        self.clusters = None
        self.histogram = []
        self.rmsTol = DEFAULT_RMSTOL # angstro        self.ignoreWaters = ignoreWaters
        self.isHydrated = False
        self.doInteractions = doInteractions
        if rmsTol:
            self.rmsTol = rmsTol
        if water_map:
            self.water_map = water_map
        #
        if auto and input_files:
            self.process()

    def __str__(self):
        text = "\n===================================================\n"
        text +=" This is my VSDockingResults object. [AutoDock]\n"
        text +=" There are many like it, but only this one\n"
        text +=" has been made from => %s\n" % self.ligName
        text +="  - receptor name            : %s\n" % self.recname
        text +="  - total runs               : %s\n" % self.totRuns
        text +="  - results poses            : %s\n" % len(self.results)
        text +="  - total clusters           : %s\n" % len(self.histogram)
        text +="  - flex res                 : %s" % self.flexres
        text +="\n Result(s) ------------------------------------------\n"
        if self.interactions:
            text += "       Energy    L.eff  Csize    Cpc   vdW      HB    Pi-Pi\n"
            for i in range(len(self.results)):
                text += " %s   %2.3f\t%2.3f\t %d\t%3.2f%%\t%d\t%d\t%d\n" % \
                    ( self.tag[i], self.results[i]['energy'], self.results[i]['leff'],
                      self.results[i]['csize'], self.results[i]['cpercent'],\
                      len( self.results[i]['vdw_contacts']),
                      (len(self.results[i]['hb']['acceptors'])+len(self.results[i]['hb']['donors'])),
                      len(self.results[i]['pi']['p-stack'])+len(self.results[i]['pi']['t-stack'])
                    )
        else:
            text += "       Energy    L.eff  Csize    Cpc\n"
            for i in range(len(self.results)):
                text += " %s   %2.3f\t%2.3f\t %d\t%3.2f%%\n" % \
                    ( self.tag[i], self.results[i]['energy'], self.results[i]['leff'],
                    self.results[i]['csize'], self.results[i]['cpercent'])

        text +=" Histogram ------------------------------------------\n"
        c = 1
        for i in self.histogram:
            text+= "   %d.  %2.3f | %d\t%s" % (c, i[0],i[1] ,"#"*i[1])
            try:
                text += " ( %s )\n" % i[2]
            except:
                text += "\n"
            c+=1
            if c == 6:
                text += "    ...   ...   ...\n"
                break
        text +="=====================================================\n"
        return text

    def process(self): # AutoDock version
        self.getPoses()
        # check here we have poses
        if len(self.poses):
            if self.DEBUG: print "found poses:", len(self.poses)
            self.FastReclustering()
            self.extractResultsPoses()
            if self.doInteractions:
                self.calcInteractions()        
        else:
            if self.DEBUG: print "[ WARNING: No poses found in DLG ! ]" 


    def appendDlg(self, file_list): # allows to add a dlg to the list TODO potentially useless
        if self.dlgList:
            for d in file_list:
                if not d in self.dlgList:
                    self.dlgList.append(d)
        else:
            self.dlgList = file_list


    def getPoses(self, include_hydrogens=False, ignoreWaters = False):
        """
        - parser of docked poses from the dlg
        - populate self.atomTypes[]
        - define self.ligName (str)
        - populate poses dictionary: { "text", "coord", "energy", 'source'}
                      ( for every pose the 'source' specify dlg containing it and the position)
        """
        #accepted_kw = [ "ATOM", "HETATM", "ROOT", "ENDROOT", "BRANCH", "ENDBRANCH", "TORSDOF", "REMARK"  ]
        accepted_kw = [ "ATOM", "HETATM", "TORSDOF"  ]
        atype_list_complete = False
        #water_found = False
        for d in self.input_files: # we expect dlgs
            inside = False
            c = 0
            ## for l in getLines(d)[50:]:
            ## XXX N3P 2012.9.20 removed to deal with new 
            ##         condenset FA@H results
            for l in getLines(d):
                if l[0:7] == "DOCKED:":
                    inside = True # used to stop parsing as soon as another keyword (i.e. "DPF> ") is found
                    l = l[8:]
                    if "MODEL" in l: # initialize pose
                        text_pose = []
                        coord = []
                        flex_res = []
                        water = []
                        c += 1
                        in_res = False
                    elif "ENDMDL" in l:
                        coord = array( coord, 'f')
                        if len(coord) == 0:
                            if DEBUG: print "[ Warning! no coordinates found in this dlg: %s ]" % (d)
                            break
                        if flex_res:
                            flex_res = getCoords(flex_res) # transform coordinates and process them
                                                                # as we do with receptor atoms   
                        # flex res atoms are added to the pose here
                        self.poses.append( { "text"         : text_pose, 
                                             "coord"        : coord, 
                                             "energy"       : e ,
                                             'source'       : (os.path.basename(d)+":"+str(c)), 
                                             "flex_res"     : flex_res,
                                             "water_bridge" : water,
                                           } ) 
                        atype_list_complete = True
                    elif "BEGIN_RES" in l:  
                        in_res = True
                        self.flexres = True # flex_res trigger
                        flex_res.append(l)
                        text_pose.append(l)
                    elif "END_RES" in l: 
                        in_res = False
                        flex_res.append(l)
                        text_pose.append(l)
                    elif l.startswith("ATOM") or l.startswith("HETATM"): # if the line is atom but *not* flexres
                        text_pose.append(l)
                        if not in_res:
                            atype = l.rsplit()[-1]
                            if (atype == 'HD' and include_hydrogens) or not ( atype in self.ignore_at):
                                try:
                                    # NOTE the flex res are not considered in the clustering
                                    coord.append([float(l[30:38]),float(l[38:46]),float(l[46:54])])
                                    if not atype_list_complete: # FUTURE: to be used for the David's reclustering method
                                        self.atomTypes.append( atype ) 
                                except:
                                    print ">WARNING! error in parsing coords in file :",d
                                    self.problematic.append(d)
                                    # the entire dlg is ignored if something bad happens
                                    # rude, but efficient on the large scale...
                                    break 
                            elif atype in self.water_types and not ignoreWaters:
                                water.append(l)
                                self.isHydrated = True
                        else:
                            flex_res.append(l)
                    elif "USER    Estimated Free Energy of Binding    =" in l:
                        e = l.split("=")[1]
                        e = float(e.split("k")[0])
                        if e == 0.:
                            if DEBUG: print "[ Warning! FEB is zero... and it shouldn't ]"
                            
                    elif l.split(None, 1)[0] in accepted_kw:
                        text_pose.append(l)
                elif l.startswith("DPF> move") and not self.ligName:
                    self.ligName = l.split("DPF> move ")[1].split(".pdbqt")[0]
                if l.startswith("DPF >") and inside: # stop reading dlg file (average lines skipped: ~30%)
                    break
            self.totRuns = len(self.poses) # TODO check this indentation?
            # Debugging printing of poses
            if self.DEBUG:
                for p in self.poses:
                    for f in p:
                        print f
                        print p[f]
                        print "==================================="
            # THERE
            #if self.isHydrated and water_auto:
            #    self.processWaters()

    # WATER
    def processWaters(self, poses = None):
        #print "...acshuly, I don't know what to do, nao..."

        self.hydro_docking_processor = processHydroDocking(gridmap=self.water_map, 
                       ignore_types = self.ignore_at, mapdistrange = 1.0,
                       conservedwaterentropy=self.fullentropy)

        #print "DEBUG>", self.ligName
        if poses == None:
            poses = self.poses

        

        for i in range(len(poses)):
            self.hydro_docking_processor.pose = poses[i]
            poses[i] = self.hydro_docking_processor.process()
            
            #poses[i] = processHydroDocking(poses[i], gridmap = self.water_map,
            #        ignore_types = self.ignore_at, mapdistrange = 1.0)
            #poses[i] = processHydroDocking(poses[i], receptor=self.receptor, gridmap = self.water_map,
             #       ignore_types = self.ignore_at, mapdistrange = 1.0)
            if self.DEBUG: print "[ CALC WATER_INTERACTIONS HERE WHEN WET WILL BE READY ]"
        #for pose in poses:
        #    for f in pose:
        #        #if "water" in f :
        #        #   for w in pose[f]:
        #        #        print w.strip()
        #        print f
        #        print pose[f]
        #print "===================="
        return


    def FastReclustering(self):
        """ self.cluster results : poses [x,y,x,z,y,z,x, ...] =>  [ [x,x,x], [y,y], [ z,z,z,z,z,z,z,z], ... ] """
        self.clusters = []
        self.poses.sort(key=lambda x: x['energy'])
        poses = self.poses[:]
        while len(poses) > 1:
            rmsdcalc = RMSDCalculator(poses[0]["coord"])
            def func(x): return rmsdcalc.computeRMSD(x["coord"]) <= self.rmsTol
            cluster = filter(func, poses[1:]) + [poses[0]]
            cluster.sort(key=lambda x: x['energy'])
            # XXX    N3P the next line has weird behaviors basing on the Numpy used...
            #        <type 'exceptions.ValueError'>: 
            #        The truth value of an array with more than one element is ambiguous. Use a.any() or a.all()
            # poses = [ p for p in poses if not p in cluster] XXX REMOVED XXX
            reminders = []
            for p in poses: # TODO change this to another function as above to use 'filter'
                used=False
                for c in cluster:
                    if p['text'] == c['text']: # XXX INEFFICIENT HERE!!
                        used = True
                        break
                if not used:
                    reminders.append(p)
            poses = reminders[:]
            self.clusters.append(cluster)
        if len(poses):
            self.clusters.append(poses)

    def extractResultsPoses(self):
        """find LE and LC (if available) and the self.histogram"""
        # TODO speed here can be improved
        LCpose = ""
        LCsize = -1
        # LCenergy = 10000. Changed for catastrophyc results with energy bigger than that
        # (i.e. too-small grid box)
        LCenergy = 1E99
        LEpose = ""
        LEsize = -1
        LEenergy = 1E99
        c = 0


        for pop in self.clusters:
            curr_csize = len(pop)
            curr_cenergy = pop[0]['energy']
            self.histogram.append([curr_cenergy, curr_csize])
            if (curr_csize > LCsize) or (curr_csize == LCsize and curr_cenergy < LCenergy ):
                LCsize = curr_csize
                LCpose = pop[0]
                LCenergy = curr_cenergy
                LCindex = c
            if curr_cenergy < LEenergy:
                LEenergy = curr_cenergy
                LEpose = pop[0]
                LEsize = curr_csize
                LEindex = c
            c += 1
        self.results.append(LEpose)
        self.results[0]['csize'] = LEsize
        self.results[0]['cpercent'] = (float(LEsize) / float(self.totRuns))*100
        self.results[0]['leff'] = self.results[0]['energy'] / float( len( self.atomTypes))
        self.results[0]['vdw_contacts'] = []
        self.results[0]['metal_coord'] = [] 
        acc, don = self.findHbAccepDon( self.results[0]['text'] )
        self.results[0]['hba_atoms'] = acc
        self.results[0]['hbd_atoms'] = don
        if self.DEBUG:
           writeList(self.ligName+"_lig_acceptors.pdb", acc)
           writeList(self.ligName+"_lig_donors.pdb", don)
        self.histogram[LEindex].append("**") # <-marker for LE
        if not LEpose == LCpose:
            self.results.append( LCpose )
            self.results[1]['csize'] = LCsize
            self.results[1]['cpercent'] =  (float(LCsize) / float(self.totRuns))*100
            self.results[1]['leff'] = self.results[1]['energy'] / float( len( self.atomTypes))
            self.results[1]['vdw_contacts'] = []
            self.results[1]['metal_coord'] = []
            acc, don = self.findHbAccepDon(self.results[1]['text'])
            self.results[1]['hba_atoms'] = acc
            self.results[1]['hbd_atoms'] = don
            self.histogram[LCindex].append("*") # <-marker for LC
        self.histogram.sort(key = lambda x: x[0])

    def generatePDBQTplus(self):
        # AutoDock version
        """ creates a formatted AutoDock PDBQT plus file 
            with all the data extracted from the re-clustering
        """
        time_info = time.localtime()[0:6]
        # pack the header:
        buff  = "USER    ADVS_result> %d-%d-%d %d:%d:%d\n"  % (time_info)
        buff += "USER    AD_rec> %s\n" % ( self.recname )
        buff += "USER    AD_runs,rmstol,tot_clusters> %d,%1.2f,%d\n" % ( self.totRuns,\
                    self.rmsTol, len(self.histogram))
        buff += "USER    AD_dlg_list> " 
        for d in self.input_files: # we expect dlgs here
            buff+="%s," % (os.path.basename(d))
        buff = buff[:-1]+"\n"
        buff += "USER    AD_results> %d\n" % ( len(self.results) ) # add the results count  (1,2)
        buff += "USER    AD_histogram> " # add the histogram
        for h in self.histogram:
            buff += "%2.3f:%d" % (h[0], h[1] )
            try:
                buff += ":"+h[2]+","
            except:
                buff += ","
        buff = buff[:-1]+"\n" # cut out the last ","
        for p in range(len(self.results)): # separate pose MODELS are handled here
            pose = self.results[p]
            buff += "MODEL   %d\n" % ( p+1)
            buff += "USER    #     energy,\tleff,\tc_size,\tc_pc\n"
            buff += "USER    AD_%s> %2.3f,\t%2.3f,\t%d,\t%3.2f\n" % (self.tag[p],\
                                    pose['energy'], pose['leff'], pose['csize'], pose['cpercent']) 
            #
            # XXX change the ligand atom numbering with the code Michel sent 3.25.2012
            #

            if self.interactions:
                if pose['hb']['acceptors']:         # add hb acceptors info
                    buff += "USER    AD_%s_hba> " % (self.tag[p])
                    for pair in pose['hb']['acceptors']:
                        buff += "%s~~%s," % (pmvAtomStrip(pair[0]), pmvAtomStrip(pair[1]))
                    buff = buff[:-1]+"\n"
                if pose['hb']['donors']:            # add hb donors info
                    buff += "USER    AD_%s_hbd> " % (self.tag[p])
                    for pair in pose['hb']['donors']:
                        buff += "%s~~%s," % (pmvAtomStrip(pair[0]), pmvAtomStrip(pair[1]))
                    buff = buff[:-1]+"\n"
                if pose['vdw_contacts']:            # add close contacts info
                    buff += "USER    AD_%s_vdw> " % (self.tag[p])
                    for a in pose['vdw_contacts']:
                        buff += "%s," % (pmvAtomStrip(a))
                    buff = buff[:-1]+"\n"
                if pose['metal_coord']:             # add metal coordination info
                    buff += "USER    AD_%s_mtl> " % (self.tag[p])
                    for pair in pose['metal_coord']:
                        buff += "%s~~%s," % (pmvAtomStrip(pair[0]), pmvAtomStrip(pair[1]))
                    buff = buff[:-1]+"\n"
                if pose['pi']['p-stack']:
                    buff += "USER    AD_%s_ppi> " % (self.tag[p])
                    for a in pose['pi']['p-stack']: # add pi interaction info (t-stack)
                        buff += "%s~~(%2.3f,%2.3f,%2.3f:%2.3f,%2.3f,%2.3f);" %\
                        (a[0], a[1][0], a[1][1], a[1][2], a[2][0], a[2][1], a[2][2])
                    buff = buff[:-1]+"\n"
                if pose['pi']['t-stack']:
                    buff += "USER    AD_%s_tpi> " % (self.tag[p])
                    for a in pose['pi']['t-stack']: # add pi interaction info (p-stack)
                        buff += "%s~~(%2.3f,%2.3f,%2.3f:%2.3f,%2.3f,%2.3f);" %\
                        (a[0], a[1][0], a[1][1], a[1][2], a[2][0], a[2][1], a[2][2])
                    buff = buff[:-1]+"\n"
            buff += "USER    AD_%s_source> %s\n" % (self.tag[p], pose['source'])
            for i in pose['text']:
                if not i.startswith("MODEL") and not i.startswith("ENDMDL"):
                    buff += i
            buff +="ENDMDL  %d\n" % ( p+1 )
        return buff
#
################################################## AutoDock class [end]

class AutoDockVinaVsResult(VSDockingResultsGenerator):
    def __init__(self, input_files = None, mode = 1, water_map = None, # XXX useless, but kept for compatibility
                recname = None, receptor = None, auto = True, 
                doInteractions = True, hbtol = 0.0, debug = False):


        VSDockingResultsGenerator.__init__(self,
            input_files = input_files,           # list of vina ligand_out.pdbqt from wich poses will be extracted
            recname = recname,         # receptor name (if not provided it will be taken from 'self.receptor'
            receptor = receptor,        # receptor coordinates: PDBQT or getCoord() output (CACHE)
            doInteractions = doInteractions,  # calculate ligand-target interactions
            hbtol = 0.0,
            debug = debug,          # debug mode
            )
        self.totRuns = 0 # Vina poses in the file
        self.mode = 1    # poses saved in the result file
        self.histogram = [] # energy | lb RMSD | ub RMSD
        if not input_files == None:
            name = os.path.basename(input_files) # the name of the ligand is guessed from the filename
            name = os.path.splitext(name)[0]  # name can be eventually changed by
                                              # overriding the self.ligName value
            self.ligName = name
            if auto:
                self.process()

    def __str__(self):
        text = "\n===================================================\n"
        text +=" This is my VSDockingResults object. [ Vina ]\n"
        text +=" There are many like it, but only this one\n"
        text +=" has been made from => %s\n" % self.ligName
        text +="  - receptor name            : %s\n" % self.recname
        text +="  - total binding modes      : %s\n" % self.totRuns
        text +="  - results saved            : %s\n" % len(self.results)
        text +="  - flex res                 : %s" % self.flexres
        text +="\n Result(s) ------------------------------------------\n"
        if self.interactions:
            text += "       Energy    L.eff  vdW      HB    Pi-Pi\n"
            for i in range(len(self.results)):
                text += "  %s   %2.3f\t%2.3f\t%d\t%d\t%d\n" % \
                    ( i+1, self.results[i]['energy'], self.results[i]['leff'],
                      len( self.results[i]['vdw_contacts']),
                      (len(self.results[i]['hb']['acceptors'])+len(self.results[i]['hb']['donors'])),
                      len(self.results[i]['pi']['p-stack'])+len(self.results[i]['pi']['t-stack'])
                    )
        else:
            text += "       Energy    L.eff\n"
            for i in range(len(self.results)):
                text += " %s   %2.3f\t%2.3f" % \
                    ( i+1, self.results[i]['energy'], self.results[i]['leff'])

        text +=" Histogram ------------------------------------------\n"
        c = 1
        text +=" Pose    E     lbRMS     ubRMS\n"
        for i in self.histogram:
            # it should be # energy | lb RMSD | ub RMSD
            text+= " %d   %2.3f |  %2.3f | %2.3f\n" % (c, i[0],i[1],i[2])
            c+=1
            if c == 9:
                text += "        ...   ...   ...\n"
                break
        text +="=====================================================\n"
        return text

    def process(self): # AutoDock Vina version
        self.getPoses()
        self.extractResultsPoses()
        if self.doInteractions:
            self.calcInteractions()

    def getPoses(self, include_hydrogens=False):
        """
        - parser of docked poses from a Vina output
        - populate self.atomTypes[]
        - populate poses dictionary: { "text", "coord", "energy", 'source'}
                      ( for every pose the 'source' specify dlg containing it and the position)
        """
        accepted_kw = [ "ATOM", "HETATM", "ROOT", "ENDROOT", "BRANCH", "ENDBRANCH", "TORSDOF", "REMARK"  ]
        atype_list_complete = False
        inside = False
        c = 0
        lines = getLines(self.input_files)
        if "REMARK VINA RESULT" in lines[1]:
            if DEBUG: print "[found Vina result pdbqt]"
            for l in lines:
                if l.strip():
                    if "MODEL" in l: # initialize pose
                        text_pose = []
                        coord = []
                        flex_res = []
                        c += 1
                        in_res = False
                    elif "ENDMDL" in l:
                        coord = array( coord, 'f')
                        if flex_res:
                            flex_res = getCoords(flex_res) # transform coordinates and process them
                                                           # as we do with receptor atoms   
                        # flex res atoms are added to the pose here
                        self.poses.append( { "text" : text_pose, "coord" : coord, "energy" : e ,\
                                'source' : (os.path.basename(self.input_files)+":"+str(c)), "flex_res" : flex_res } ) 
                        atype_list_complete = True
                    elif "BEGIN_RES" in l:  
                        in_res = True
                        self.flexres = True # flex_res trigger
                        flex_res.append(l)
                        text_pose.append(l)
                    elif "END_RES" in l: 
                        in_res = False
                        flex_res.append(l)
                        text_pose.append(l)
                    elif l.startswith("ATOM") or l.startswith("HETATM"):
                        text_pose.append(l)
                        if not in_res:
                            atype = l.rsplit()[-1]
                            if (not atype == "HD") or include_hydrogens : # HD should be excluded from array/RMSD calculation
                                try:
                                    coord.append([float(l[30:38]),float(l[38:46]),float(l[46:54])])
                                    if not atype_list_complete: # FUTURE: to be used for the David's reclustering method
                                        self.atomTypes.append( atype ) 
                                except:
                                    print ">WARNING! error in parsing coords in file :",self.input_files
                                    self.problematic.append(self.input_files)
                                    break 
                        else:
                            flex_res.append(l)
                    elif "REMARK VINA RESULT:" in l:
                        values = l.split(":")[1]
                        values = values.split()
                        e = float(values[0])
                        lbrms = float(values[1])
                        ubrms = float(values[2])
                        self.histogram.append([e, lbrms, ubrms])
                    elif l.split(None, 1)[0] in accepted_kw:
                        text_pose.append(l)
            self.totRuns = len(self.poses)
        else:
            print ">WARNING! not a Vina output PDBQT file :",self.input_files
    
    def extractResultsPoses(self):
        """extract the best result (the first?) accordingly to Vina"""
        # TODO figure out if we want some more from the result?
        # so far it is good in this way...
        if self.mode > self.totRuns:
            self.mode = self.totRuns
        for i in range(0, self.mode):
            self.results.append(self.poses[i])
            self.results[i]['leff'] = self.results[i]['energy'] / float( len( self.atomTypes))
            self.results[i]['vdw_contacts'] = []
            self.results[i]['metal_coord'] = [] # TODO 
            acc, don = self.findHbAccepDon( self.results[i]['text'] )
            self.results[i]['hba_atoms'] = acc
            self.results[i]['hbd_atoms'] = don
        if self.DEBUG:
           writeList(self.ligName+"_lig_acceptors.pdb", acc)
           writeList(self.ligName+"_lig_donors.pdb", don)

    def generatePDBQTplus(self):
        # Vina version
        """ creates a formatted AutoDock PDBQT plus file 
            with all the data extracted from the re-clustering
        """
        time_info = time.localtime()[0:6]
        # pack the header:
        buff  = "USER    ADVS_Vina_result> %d-%d-%d %d:%d:%d\n"  % (time_info)
        buff += "USER    ADVina_rec> %s\n" % ( self.recname )
        buff += "USER    ADVina_poses> %d\n" % ( self.totRuns)
        buff += "USER    ADVina_input_file> %s\n" % (os.path.basename(self.input_files))
        buff += "USER    ADVina_results> %d\n" % ( len(self.results) ) # add the results count  (1,2,...)
        buff += "USER    ADVina_histogram> " # add the histogram
        for h in self.histogram:
            buff += "%2.3f:%2.3f:%2.3f," % (h[0], h[1], h[2] )
        buff = buff[:-1]+"\n" # cut out the last ","
        for p in range(len(self.results)): # separate pose MODELS are handled here
            pose = self.results[p]
            buff += "MODEL   %d\n" % ( p+1)
            buff += "USER    #     energy,\tleff\n"
            buff += "USER    ADVina_pose%d> %2.3f,\t%2.3f\n" % (p+1,\
                                    pose['energy'], pose['leff'])
            if self.interactions:
                if pose['hb']['acceptors']:     # add hb acceptors info
                    buff += "USER    ADVina_pose%d_hba> " % (p+1)
                    for pair in pose['hb']['acceptors']:
                        buff += "%s~~%s," % (pmvAtomStrip(pair[0]), pmvAtomStrip(pair[1]))
                    buff = buff[:-1]+"\n"
                if pose['hb']['donors']:        # add hb donors info
                    buff += "USER    ADVina_pose%d_hbd> " % (p+1)
                    for pair in pose['hb']['donors']:
                        buff += "%s~~%s," % (pmvAtomStrip(pair[0]), pmvAtomStrip(pair[1]))
                    buff = buff[:-1]+"\n"
                if pose['vdw_contacts']:        # add close contacts info
                    buff += "USER    ADVina_pose%s_vdw> " % (p+1)
                    for a in pose['vdw_contacts']:
                        buff += "%s," % (pmvAtomStrip(a))
                    buff = buff[:-1]+"\n"
                if pose['metal_coord']:         # add metal coordination info
                    buff += "USER    ADVina_pose%s_mtl> " % (p+1)
                    for pair in pose['metal_coord']:
                        buff += "%s~~%s," % (pmvAtomStrip(pair[0]), pmvAtomStrip(pair[1]))
                    buff = buff[:-1]+"\n"   
                if pose['pi']['p-stack']:       # add pi interaction info
                    buff += "USER    ADVina_pose%s_ppi> " % (p+1)
                    for a in pose['pi']['p-stack']:
                        buff += "%s~~(%2.3f,%2.3f,%2.3f:%2.3f,%2.3f,%2.3f);" %\
                        (a[0], a[1][0], a[1][1], a[1][2], a[2][0], a[2][1], a[2][2])
                    buff = buff[:-1]+"\n"
                if pose['pi']['t-stack']:       # add pi interaction info
                    buff += "USER    ADVina_pose%s_tpi> " % (p+1)
                    for a in pose['pi']['t-stack']:
                        buff += "%s~~(%2.3f,%2.3f,%2.3f:%2.3f,%2.3f,%2.3f);" %\
                        (a[0], a[1][0], a[1][1], a[1][2], a[2][0], a[2][1], a[2][2])
                    buff = buff[:-1]+"\n"
            buff += "USER    ADVina_pose%s_source> %s\n" % (p+1, pose['source'])
            for i in pose['text']:
                if not i.startswith("MODEL") and not i.startswith("ENDMDL"):
                    buff += i
            buff +="ENDMDL  %d\n" % ( p+1 )
        return buff
###
###    def getInteractSummary(self, pose):
###        """ collect nd format the interaction information for the pose"""
###        interactions = {}
###        if pose.hasattrib('hba'):
###            interactions['hba'] = [ pmvAtomStrip(p[0]), pmvAtomStrip(p[1]) for p in pose['hb']['acceptors'] ]
###        if pose.hasattrib('hbd'):
###            interactions['hbd'] = [ pmvAtomStrip(p[0]), pmvAtomStrip(p[1]) for p in pose['hb']['donors'] ]
###        if pose.hasattrib['vdw']:
###            interactions['vdw'] = [ pmvAtomStrip(p) for p in pose['vdw_contacts'] ]
###        if pose.hasattrib['metal']:
###            interactions['metal'] = [ pmvAtomStrip(p[0]), pmvAtomStrip(p[1]) for p in pose['metal_coord'] ]
###        if pose.hasattrib['ppi']:
###            interactions['ppi'] = [ pose['pi']['p-stack'] ]
###        if pose.hasattrib['tpi']:
###            interactions['tpi'] = [ pose['pi']['t-stack'] ]
###        return interactions
###
###





#       
#           AutoDock | Raccoon2
#
#       Copyright 2013, Stefano Forli
#          Molecular Graphics Lab
#  
#     The Scripps Research Institute 
#           _  
#          (,)  T  h e
#         _/
#        (.)    S  c r i p p s
#          \_
#          (,)  R  e s e a r c h
#         ./  
#        ( )    I  n s t i t u t e
#         '
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#!/usr/bin/env python
            
import AutoDockTools.HelperFunctionsN3P as hf
import RaccoonSshTools as Rst
import RaccoonServices 
import RaccoonLibraries 
import sys
import json
from DebugTools import DebugObj
import threading 
import zlib
import operator
import urllib
import os
import tarfile
from datetime import datetime

# Laws are like sausages. It's better not to see them being made.
# Otto von Bismarck
# German Prussian politician (1815 - 1898)  


class RaccoonRemoteServer(DebugObj): 
    # XXX TODO:
    # add support for system and user paths
    # RACDIR_USER, RACDIR_SYSTEM
    # 
    # -add encryption for passwords
    #   http://www.blog.pythonlibrary.org/2010/10/20/downloading-encrypted-and-compressed-files-with-python/

    """
        RaccoonServer( address,             |
                       username,            |    Server connection data
                       password=None,       |    password is optional if keys are available 
                       usesyshostkeys=1     |--- in the user .ssh/ directory
                       autoaddmissingkeys(0)|    by default missing keys are added automatically
                       pkey=None            |    pkey='' specifies a public key file

                       name = ''            |--- Name of the RaccoonServer

                       racdir = None        |--- Set the initial path where to look for the 
                                                 config files. This value can be defined on
                                                 the server by setting the environmental
                                                 variable $RACDIR
                                                 (i.e. Bash: "export RACDIR='~/racdir/')

                       autoconnect (1)      |--- Open a connection when initialized

                       check       (1)      |___ check if the server is 'racconized'
                                            |    (i.e. necessary dirs are in place)
                                            
                       prepare     (0)      |---- racconize the server when initialized

                       autoscan    (1)      |---- autoscan=True 

                       debug       (0)      |---- activate debug mode (verbose)
                      )

    Supported systems are any *NIX system with a PBS or SGE scheduler.

    This is the structure of a racconized remote server:

    $RACDIR/ ***
        |
        +- bin/  store binaries installed by services
        |
        +- tmp/  ( store temporary stuff for calculations, like config files, etc)
        |
        +- services/  
        |     |          
        |     +- $(service1).conf ( service1 specific settings ) 
        |     |
        |     +- $(service2).conf ( service2 specific settings )
        |
        +- libraries/ ***
        |     |
        |     +- index.lib *** ( list of libraries+type(lig/rec/...) + data(properties,items,...))
        |     |
        |     +- $(libname1)/   [ OPTIONAL: if any in 'index.lib' ]
        |           |
        |           +- library.db    [ sqlite3|text|json ? ]
        |           |
        |           + 00
        |           |  |
        |           |  +-- mol001.pdbqt
        |           |  '-- mol002.pdbqt
        |           |
        |           + 01
        |           |
        |           ...
        |
        +- data/  ***
        |    |
        |    +- history.log  ***  ( history and status of jobs  SQLITE? JSON?)
        |    |
        |    +- $(outname1) -+  [ OPTIONAL: if 'outname1' listed in 'history.log' ]
        ...

    When racconizing a server, the following files will be created: 
     
        UPDATE!
     basic directories:
         $RACDIR, $RACDIR/config/, $RACDIR/libraries, $RACDIR/data

     basic files:
         $RACDIR/libraries/index.lib, $RACDIR/data/history.log

    """


    def __init__(self, address=None,                    # server address
                       username=None,                   # username MANDATORY
                       password=None,                   # password (optional if keys used and available)
                       name='RaccoonSsh_DockingServer', # name of the Raccoon Server
                       debug=False,                     # debug verbose mode
                       usesyshostkeys=True,             # try to connect without password using locally stored keys (POSIX OS ONLY)
                       autoaddmissingkeys=True,         # policy for adding the server to known hosts at first connection 
                       pkey=None,                       # filename of the encryption key to use
                       autoconnect=False,               # open connection with the server after initialization
                       check = True,                    # check if the server is racconized
                       prepare=False,                   # racconize the server if necessary
                       autoscan=True,                   # scan for services and resources upon connection
                       racdir=None,                     # use specified remote path as $RACDIR
                       scratchdir=None ):               # use local tempdir as scratch space (for services that need to write files)

        # debugging tools
        DebugObj.__init__(self, debug)

        # ssh client
        self.ssh = Rst.RaccoonSshClient(address, username, password, name, debug, 
                usesyshostkeys, autoaddmissingkeys, pkey, autoconnect)


        # file transfer (multi-threadable)
        self.transfer = None

        ### # on-going file transfers are stored here
        ### self._pending_transfers = []

        # DEFAULTS
        # base path for all the config, settings, data files 
        RACDIR = '~/raccoon'
        SCRATCH = os.path.expanduser('~') + os.sep + '.raccoon_tmp'
        
        # local scratch dir
        if not scratchdir == None:
            self.scratchdir = scratchdir
        else:
            self.scratchdir = SCRATCH

        # test if local scratch dir exists
        if not os.path.exists(self.scratchdir):
            if not hf.makeDir(fullpath=self.scratchdir):
                print "SERIOUS TROUBLES AHEAD: ERRORS CREATING TEMPDIR [%s]"  % self.scratchdir

        # Raccoon server specific properties

        # it is possible to connected to the server
        self.properties = { 'name': name, 'online' : False, 'hostname' : ''}
                          
        # server is raccoon-compliant (raccoonized)
        self.properties['ready'] = False

        # system info
        self.properties['system'] = { 'architecture' : None, # uname -m
                                      'scheduler'    : None, # pbs, sge 
                                      'os'           : None, # uname -o
                                      'kernel'       : None, # uname -r
                                      'hostname'     : None,
                                      'nodes'        : None,
                                      'scheduler_info': []  # populated after scheduler identification
                                    }
        # $RACDIR 
        if racdir:
            self.setRacPath(racdir)
        else:
            self.setRacPath(RACDIR)

        # this defines the raccoon-type, and it can be one of more of 'docking', 'md', 'docking_postprocess', ...
        self.properties['services'] =  {}

        # libraries available on the server (read from the Library Index file)
        # the content of the library index is defined by RaccoonLibraryManager.LibraryObj
        self.properties['libraries'] = [] 

        # list of unknown services found on the remote server (from config files)
        self._unknown_services = [] # = ( [ service_type, config_fname ] )
        # list of unknown library types found on the server (from index files)
        self._unknown_libs = [] # = ( lib_dict )

        # history of jobs submitted to the server
        self.history = {}

        if autoconnect:
            self.dprint("auto-connect requested. Opening connection")
            #conn = self.ssh.openconnection()
            conn = self.connect()
            if conn == 1:
                # check if the server is racconized already
                if check:
                    status = self.checkServerStatus()
                    self.dprint('auto-check requested. Status[%s]' % status)
                    if status:
                        # check if there are services configured on the server
                        if autoscan:
                            self.dprint('auto-scan requested. Scanning for services on the server...')
                            self.scan()
                # check if it needs to be prepared and it was requested so
                if prepare:
                    self.dprint('auto-prepare requested.')
                    self.racconize(force=False)
            else:
                self.dprint("connection not succesful [%s]" % conn)

    def hostname(self):
        """ return the host name as known by the host itself"""
        return self.properties['hostname']

    def name(self):
        """ return server name """
        return self.properties['name']

    def date(self):
        """ simple date function provided for services timestamps"""
        d = datetime.now()
        return [d.year, d.month, d.day, d.hour, d.minute, d.second ]

    def connect(self):
        """used to start connection from the external"""
        self.dprint("connection requested")
        conn = self.ssh.openconnection()
        if conn == 1:
            self.dprint("setting the mule transfer object") 
            self.transfer = Rst.FileMule( ssh = self.ssh, debug = self.debug)
            self.getHostname()
        else:
            self.dprint("disabling the mule transfer object") 
            self.transfer = None
            self.properties['hostname'] = ""
        return conn

    def is_connected(self):
        """ check if there is a working connection already open"""
        if not self.ssh.connection == None:
            return True
        return False

    def getHostname(self):
        """ query the server for its own name"""
        cmd = """hostname"""
        report = self.ssh.execmd(cmd)
        self.dprint("querying the server for hostname...")
        if report == False:
            self.dprint("error in connection")
            return False
        stdout, stderr = report
        if len(stdout):
            self.properties['hostname'] = stdout[0]
            return True
        return False

    def disconnect(self):
        """close open connections and disable transer object"""
        if not self.transfer == None and self.transfer._running:
            self.dprint("running transfer detected... stopping it")
            self.transfer.stop()
        self.transfer = None
        self.ssh.closeconnection()

    def scan(self):
        """ scan the server for libraries and services, and read history file"""
        if self.checkServerStatus():
            self.findLibraries()
            self.findServices()
            self.loadHistory()
        else:
            self.dprint("The server status is not true, returning")
            return
        
    def writeJson(self, fname, data, mode='w', compression=False):
        """ write a python data type (dict or list) 
            into a JSON file

            optional Zlib compression available
            
            # COMPRESSION
            http://www.blog.pythonlibrary.org/2010/10/20/\
                downloading-encrypted-and-compressed-files-with-python/
        """
        self.dprint("Open file [%s] for writing, mode [%s]" % (fname, mode))
        #report = self.ssh.openfile(fname, mode=mode)
        report = self.ssh.openfile(fname, mode='w')
        self.dprint("data type is [%s]" % type(data))
        try:
            fp, sftp = report
            if not isinstance(report, tuple):
                self.dprint("ERROR: file pointer not generated: [%s]" % report)
                exit(1)
        except:
            self.dprint("Error writing JSON file [%s] : [%s]" % (fname, report))
            #fp.close()
            #sftp.close()
            return False
        if compression:
            self.dprint("writing (COMPRESSED) to [%s]" % fp)
            data = json.dumps(data, indent=2)
            zdata = zlib.compress(data)
            fp.write(zdata)
        else:
            self.dprint("writing (UNCOMPRESSED) to [%s]" % fp)
            json.dump(data, fp, indent=2)
        self.dprint("Data written")
        fp.close()
        sftp.close()
        return True


    def readJson(self, fname, compression = False):
        """ read a json data file and perform the 
            opportune conversions to restore python types
            
            supported types are dict and list

            optional Zlib compression available

            # source:
            # http://stackoverflow.com/questions/956867/how-to-get-string-objects-instead-unicode-ones-from-json-in-python
        """
        self.dprint("Open file for reading [%s]" % (fname))
        report = self.ssh.openfile(fname, mode='r')
        try:
            fp, sftp = report
        except:
            self.dprint("Error reading JSON file [%s] : [%]" % (fname, report))
            return False
        
        data = fp.read()
        fp.close()
        sftp.close()
        if compression:
            data = zlib.decompress(data)
        try:
            data = json.loads(data, object_hook=hf.jsonhelper)
        except:
            self.dprint("Problem [%s]" % sys.exc_info()[1])
            return False
        self.dprint("data type read is [%s]" % type(data))
        return data


    def writetext(self, fname, text):
        """ write a text file in specified location"""
        self.dprint("writing text file [%s]" % fname)
        try:
            fp, sftp = self.ssh.openfile(fname, mode='w')
            fp.write(text)
            fp.close()
            sftp.close()
            return True
        except:
            error = sys.exc_info()[1]
            self.dprint(error)
            return error


    def setRacPath(self, path):
        """ set the RacDir path variable and configure
            all depending variables basing
        """
        CONFIG_PATH = 'services'
        LIBRARY_PATH = 'libraries'
        LIBRARY_FILE = 'index.lib'
        DATA_PATH = 'data'
        BIN_PATH = 'bin'
        TMP_PATH = 'tmp'
        HISTORY_FILE = 'history.log'

        self.properties['racdir'] = path + '/'
        # config directory name ( $RACDIR/'services' )
        self.properties['config_path'] = self.properties['racdir'] + CONFIG_PATH + '/'
        # bin path
        self.properties['bin_path'] = self.properties['racdir'] + BIN_PATH + '/'
        # temp path
        self.properties['temp_path'] = self.properties['racdir'] + TMP_PATH + '/'
        # data path
        self.properties['data_path'] = self.properties['racdir'] + DATA_PATH + '/'
        # libraries path
        self.properties['library_path'] = self.properties['racdir'] + LIBRARY_PATH + '/'
        # libraries index file
        self.properties['library_index'] = self.properties['library_path'] + LIBRARY_FILE
        # history file
        self.properties['history'] = self.properties['data_path'] + HISTORY_FILE


    def getRacPath(self):
        """ return the current RACDIR path 
            test if there is environment variable RACDIR 
            set, or else use the default value
        """
        # XXX THIS FUNCTION MUST BE UPDATED TO 
        # XXX HANDLE SYSTEM AND USER DIRS TOGETHER
        cmd = 'printenv | grep RACDIR'
        report = self.ssh.execmd(cmd)
        if report == False:
            print "getRacPath> No connection"
            return
        stdout, stderr = report
        if len(stdout):
            found = stdout[0]
            if not found == '':
                racdir = found.split("=")[1]
                self.dprint("Found remote env variable $RACDIR on the server [%s]"% racdir)
                self.properties['racdir']
        return self.properties['racdir']

    def getConfigPath(self):
        """ return the path containing the services config files"""
        return self.properties['config_path']

    def getLibraryPath(self):
        """ return the library path """
        return self.properties['library_path']

    def getLibraryIndexFilename(self):
        """return the library index file"""
        return self.properties['library_index']

    def getTempPath(self):
        """return the path to the temp directory"""
        return self.properties['temp_path']

    def getBinPath(self):
        """return the path to the bin directory"""
        return self.properties['bin_path']

    # data directory
    def getDataPath(self):
        """ return the path where data is stored """
        return self.properties['data_path']

    # history stuff
    def getHistoryFilename(self):
        """ return the history of jobs run on the server"""
        return self.properties['history']
        
    def loadHistory(self):
        """ return the history of jobs submitted to the server
        """
        fname = self.getHistoryFilename()
        self.dprint("reading [%s] history file" % fname)
        self.history = self.readJson(fname)
        if self.history == False:
            self.dprint("empty/unreadable history file")
            self.history = {}

    def saveHistory(self):
        """ return the history of jobs run on the server
        """
        fname = self.getHistoryFilename()
        self.dprint("writing [%s] history file" % fname)
        print "This is the history file to write:", fname
        return self.writeJson(fname, self.history)

        
    def loadLibraryIndex(self):
        """ read and parse the master Library Index file
            (JSON) containing libraries installed on the server
            and return the library index (DICT)

            the content of the library index is defined by 
            RaccoonLibraryManager.LibraryObj
        """
        fname = self.getLibraryIndexFilename()
        self.dprint("reading library index file [%s]" % fname, new=1)
        libindex = self.readJson(fname)
        print "\n\n\n\n--------------------------------"
        print libindex
        print "--------------------------------\n\n\n"
        #libraries = []
        del self.properties['libraries'][:]
        del self._unknown_libs[:]
        if libindex == False:
            self.dprint("error library index")
            return False
        unknonw_lib = []
        for data in libindex:
            lib = RaccoonLibraries.knownlibraries.get(data['type'], None)
            if lib:
                print "LIBINDEX FOUND>>>> ", data['index_file']
                lib = lib(server = self, info = data) 
                self.properties['libraries'].append(lib)
            else:
                self._unknown_libs.append(data)
        self.properties['libraries'] = sorted(self.properties['libraries'],
                                        key = operator.methodcaller('name') )

        print self.properties['libraries']

    def saveLibraryIndex(self):
        """ write the list of libraries installed on the server 
            in the master Library Index file

            the content of the library index is defined by 
            RaccoonLibraryManager.LibraryObj
        """
        fname = self.getLibraryIndexFilename()
        self.dprint("writing library index file [%s]" % fname, new=1)
        print  self.properties['libraries'], type( self.properties['libraries'])
        data = [ x.info for x in self.properties['libraries'] ]
        report = self.writeJson(fname, data)
        if report == False:
            self.dprint("library index writing failed")
        self.dprint("library index writing successful")
        return report

    def delLibrary(self, name, delfiles=True):
        """ unregister a library and update the library index """
        lib = self.getLibrary(name)
        if lib:
            if delfiles:
                self.dprint('deleting files...')
                path = lib.getFilesPath()
                self.ssh.delfiles([path])
            self.properties['libraries'].remove(lib)
            self.saveLibraryIndex()
            return True
        else:
            self.dprint('no library with name [%s]' % name)
            return False

    def getLibraries(self):
        """ return libraries list"""
        return self.properties['libraries']

    def getLibrary(self, name):
        """ return a library with name"""
        for l in self.getLibraries():
            if l.name() == name:
                return l
        return None
            
    def getLibraryNames(self):
        """ return list of library names"""
        return [ l.name() for l in self.getLibraries() ]

    
###    def findJob(self, name, prj=None, exp=None,  _type='vs'):
###        """ check if jobname has already been found in the currently loaded history file"""
###        self.dprint("searching for [%s|%s|%s](%s)" % (prj, exp, name, _type) )
###        if prj == None:
###            prj_pool = self.history.keys()
###        else:
###            prj_pool = [prj]
###        if prj in prj_pool:
###            if not prj in self.history.keys():
###                self.dprint("*NOT FOUND* (no projects with this name)") 
###                return ()
###            if exp == None:
###                exp_pool = self.history[prj].keys()
###            else:
###                exp_pool = [exp]
###            if not exp in self.history[prj].keys():
###                self.dprint("*NOT FOUND* (no experiments with this name)") 
###                return ()
###            for exp in exp_pool:
###                if exp in self.history[prj].keys():
###                    if name in self.history[prj][exp].keys():
###                        found = ( prj, exp, name ) 
###                        self.dprint("*FOUND* [%s|%s|%s]" % found)
###                        return found
###        self.dprint("*NOT FOUND*") 
###        return ()

    def findJob(self, name, prj=None, exp=None): #,  _type='vs'):
        """ check if jobname has already been found in the currently loaded history file"""
        self.dprint("searching for [%s|%s|%s]" % (prj, exp, name) )#, _type) )
        if prj == None:
            prj_pool = self.history.keys()
        else:
            prj_pool = [prj]
        self.dprint("PRJ POOL [%s]" % ",".join(prj_pool) )
        for p in prj_pool:
            if not p in self.history.keys():
                self.dprint("*NOT FOUND* (no projects with this name)") 
                return ()
            if exp == None:
                exp_pool = self.history[p].keys()
            else:
                exp_pool = [exp]
            self.dprint("\nEXP POOL [%s]" % ",".join(exp_pool) )
            for e in exp_pool:
                if not e in self.history[p].keys():
                    self.dprint("*NOT FOUND* (no experiments with this name)") 
                    return ()
                else:
                    if name in self.history[p][e].keys():
                        found = ( p, e, name ) 
                        self.dprint("\n*FOUND* [%s|%s|%s]" % found)
                        return found
        self.dprint("*NOT FOUND*") 
        return ()



    def getJobData(self, jobdata): # prj=None, exp=None, name=None):
        """ return info about a job prj|exp|name"""
        print "CALLED WITH THIS DATA", jobdata
        prj, exp, name = jobdata
        self.dprint("requested [%s|%s|%s]" % (prj, exp, name) )
        print "requested [%s|%s|%s]" % (prj, exp, name)
        if prj == exp == name == None:
            return ()
        try:
            self.dprint("found job")
            return self.history[prj][exp][name]
        except:
            self.dprint("no jobs")
            return ()

    def makeExperimentDir(self, job_info={}, _type='vs'):
        """ create the remote directory where the calculation 
            is going to be performed
            _type : set the type of calculation to be performed
                    so it adds an extra dir for the specific type (i.e. "vs")
        """
        base = self.getDataPath()
        self.dprint("base data path [%s], experiment type[%s]" % (base, _type))
        prj = job_info['prj']
        exp = job_info['exp']
        tag = job_info['tag']
        # makedir
        maindir = '/'.join([base, prj, exp, _type]) # $RACDIR/data/project/exp/_type
        #maindir = maindir.replace(' ', '_')
        maindir = "".join(hf.validPath(maindir))
        self.dprint("creating dir [%s]" % maindir)
        status = self.ssh.makedir(maindir)
        if status ==True:
            return True, maindir
        else:
            return False, status


    def manifesto(self, _format='json'):
        """ collects the information about 
            the server available so far.

        _format can be 'json' or 'none' ( = python dict)

        """
        p = self.properties
        t = self.serviceTypes()
        #s = self.services()
        manifesto = {}
        # sys info = p['system']
        manifesto = {'name': p['name'],
                     'services' : {},
                     'host_info': p['system'],
                     'libraries' : {},
                    }
        skip_kw = set([ 'service', 'command', 'file', 'validated', 'file',
                        'upload', 'outdir',
                        ])
        self.dprint("collecting data from the known services", args=True)
        for _type in t:
            manifesto['services'][_type] = []
            for service in self.services(_type):
                info = {}
                s = service.config
                for kw in set(s.keys())-skip_kw:
                    info[kw] = s[kw]
                manifesto['services'][_type].append(info)


        libs = self.properties['libraries']
        lib_kw = [ 'comments', 'count', 'date', 'format', 'type' ]
        if len(libs) > 0:
            self.dprint("[%d] libraries found" % len(libs))
            for l in libs:
                name = l.info['name']
                info = {}
                for kw in lib_kw:
                    info[kw] = l.info[kw]
                manifesto['libraries'][name] = info
        else:
            self.dprint("no libraries found")

        if _format == 'json':
            manifesto = json.dumps(manifesto, indent=3, sort_keys=True)
            self.dprint("JSON manifesto requested:\n\n %s\n" % manifesto)
        return manifesto


    def systemInfo(self, key=None):
        """ return specific system info (i.e. key = 'scheduler'),
            or the entire info dictionary collected so far
        """
        if key:
            return self.properties['system'].get(key, 'unknown')
        return self.properties['system']


    def isServerAccessible(self):
        """ check if it is possible to open a connection to the 
            server and update the self.properties['online' status
        """
        self.dprint("attempting opening a connection")
        status = self.ssh.openconnection()
        if not status == 1:
            self.dprint("error opening the connection", status)
            print "isServerAccessible> ERROR: %s"% status
        else:
            self.dprint("server replied and is online!")
            self.properties['online'] = True

    def getSystemInfo(self):
        """gather information on the server
           and set values in self.properties['system']
        """
        # get the hostname
        self._getHostname()
        # find which architecture is available
        self._archInfo()
        # find which scheduler(s?) are available, if any
        self._schedulerInfo()

    def _getHostname(self):
        """ find and store server hostname """
        self.properties['system']['hostname'] = ""
        cmd = """hostname"""
        report = self.ssh.execmd(cmd)
        if report == False:
            print "_getHostname> connection error"
            return
        stdout, stderr = report
        self.properties['system']['hostname'] = stdout[0]
        
    def _archInfo(self):
        """ gather general information about the system 
            hardware/software
        """
        # arch -> x86_64(garibaldi,atara), ppc(esau)
        uname_opt = [ ['architecture', '-m'],
                      ['os', '-o'],
                      ['kernel', '-r']
                    ]
        for kw, opt in uname_opt:
            self.properties['system'][kw] = None
            cmd = 'uname %s' % opt
            report = self.ssh.execmd(cmd)
            if report == False:
                print "_archInfo> No connection"
                return False
            stdout = report[0]
            stderr = report[1]
            if not len(stderr)>0 and len(stdout):
                self.properties['system'][kw] = stdout[0].strip()
            else:
                print "_archInfo> ERROR REPORT [%s][%s]" % (kw, opt), report

    def _schedulerInfo(self):
        """ try to guess which scheduler is installed on the
            remote cluster

            supported schedulers are pbs and sge
        """
        self.properties['system']['scheduler'] = '?'
        # PBS
        if not self._testPbs():
            self._testSge()
        

    def _testPbs(self):
        """ try to identify the SGE scheduler on the cluster
        
               # PBS attributes (max job arrays) : http://linux.die.net/man/7/pbs_server_attributes
        # http://www.usc.edu/hpcc/pbsman/man7/pbs_server_attributes.html 
        
        """
        found = False
        self.dprint("Testing for PBS..."),
        # XXX Better using env? 
        # cmd = """printenv | grep PBS ..."""
        cmd = """man 1 qsub >x ; grep -iq pbs x ; echo $?; rm x"""
        report = self.ssh.execmd(cmd)
        if report == False:
            print "_schedulerInfo> No connection"
            return False
        stdout = report[0]
        if len(stdout):
            stdout = stdout[0]
        self.dprint("STDOUT[%s]" % stdout)
        if str(stdout) == '0':
            found = True
        else:
            return False
        self.properties['system']['scheduler'] = 'pbs'
        self.properties['system']['nodes'] = '?'
        self.dprint("Found possible PBS scheduler")

        self.dprint("checking nodes..."),
        cmd = """pbsnodes -l all | wc -l"""
        report = self.ssh.execmd(cmd)
        if report == False:
            return False
        try:
            stdout = report[0][0]
            self.dprint("found [%s]" % stdout)
            self.properties['system']['nodes'] = int(stdout)
        except:
            self.dprint("Error [%s]" % sys.exc_info()[1])
            return False
        # get general scheduler info
        cmd = '''qmgr -c "list server"'''
        report = self.ssh.execmd(cmd)
        if report == False:
            self.dprint("WARNING: problem communicating with the PBS server (possible troubles ahead)")
            return True
        stdout, stderr = report

        self.properties['system']['scheduler_info'] = {}
        settings = [ ['max_job_array_size', int],
                     ['default_queue', str],
                     ['resources_assigned.ncpus', int],
                     ['resources_assigned.nodect', int],
                     ['default_queue', str],
                   ]
        for line in stdout:
            for kw, fn in settings:
                if kw in line:
                    v = line.split('=')[1].strip()
                    self.properties['system']['scheduler_info'][kw] = fn(v)
        return True


    def _testSge(self):
        """ try to identify the SGE scheduler on the cluster"""

        # SGE
        self.dprint("Testing for SGE..."),
        # XXX Better using env? 
        # cmd = """printenv | grep SGE ..."""
        cmd = """man 1 qsub  > x;  grep -qi "Grid Engine"  x ; echo $?; rm x"""
        report = self.ssh.execmd(cmd)
        if report == False:
            print "_schedulerInfo> No connection"
            return False
        stdout = report[0]
        self.dprint("STDOUT[%s]" % stdout)
        if len(stdout):
            stdout = stdout[0]
        if str(stdout) == '0':
            self.properties['system']['scheduler'] = 'sge'
            self.properties['system']['nodes'] = '?'
            self.dprint("Found possible SGE scheduler")
        else:
            self.dprint("No known schedulers found on the server.")
            return False
        self.dprint("checking nodes..."),
        cmd = """qhost | wc -l"""
        report = self.ssh.execmd(cmd)
        if report == False:
            return False
        self.properties['system']['scheduler_info'] = {}
        try:
            stdout = int(report[0][0])
            self.dprint("found [%s]... minus 3 lines = %s " % (stdout, stdout-3 ))
            self.properties['system']['nodes'] = stdout-3
            #self.properties['system']['scheduler_info'] = ['not available yet, TODO']
        except:
            self.dprint("Error [%s]" % sys.exc_info()[1])
            self.properties['system']['nodes'] = "(error)"
        # max jobs per array 
        cmd = """qconf -sconf | grep max_aj_instances | awk '{print $2}'"""
        report = self.ssh.execmd(cmd)
        if report == False:
            return False
        try:
            stdout = int(report[0][0])
            self.dprint("found [%s]... max_aj_instances" % (stdout))
            self.properties['system']['scheduler_info']['max_aj_instances'] = stdout
        except:
            self.dprint("'max_aj_instances' not found")
        return True

    def getScheduler(self):
        """ return the scheduler found on the cluster"""
        return self.properties['system']['scheduler']

    def getMaxJobsPerArray(self):
        """ return the maximum number of jobs that
            a job array can spawn
        """
        sched = self.getScheduler()
        if sched == 'pbs':
            kw = 'max_job_array_size'
        elif sched == 'sge':
            kw = 'max_aj_instances'
        value = self.properties['system']['scheduler_info'].get(kw, None)
        return value

    def checkRaccoonDir(self):
        """ check if the raccoon dir exists on the server """
        if not self.ssh.exists(self.getRacPath()):
            self.dprint("missing Raccoon dir [%s|%s]" % (self.getRacPath(), 
                self.ssh.normpath(self.getRacPath() )))
            return False
        return True

    def checkConfigPath(self):
        """ check if the config path exists on the server """
        if not self.ssh.exists(self.getConfigPath()):
            self.dprint("missing Config dir [%s|%s]" % (self.getConfigPath(),
                self.ssh.normpath(self.getConfigPath() )))
            return False
        return True

    def checkLibraryPath(self):
        """ check if the library path exists on the server """
        if not self.ssh.exists(self.getLibraryPath()):
            self.dprint("missing Library dir [%s|%s]" % (self.getLibraryPath(), 
            self.ssh.normpath(self.getLibraryPath() )))
            return False
        return True

    def checkDataPath(self):
        """ check if data path exists on the server """
        if not self.ssh.exists(self.getDataPath()):
            self.dprint("missing Data dir [%s|%s]" % (self.getDataPath(), 
            self.ssh.normpath(self.getDataPath() )))
            return False
        return True

    def checkTempPath(self):
        """ check if temp path exists on the server """
        if not self.ssh.exists(self.getTempPath()):
            self.dprint("missing Temp dir [%s|%s]" % (self.getTempPath(), 
            self.ssh.normpath(self.getTempPath() )))
            return False
        return True


    def checkBinPath(self):
        """ check if bin path exists on the server """
        if not self.ssh.exists(self.getBinPath()):
            self.dprint("missing Bin dir [%s|%s]" % (self.getBinPath(), 
            self.ssh.normpath(self.getBinPath() )))
            return False
        return True


    def checkLibraryIndex(self):
        """ check if library index file exists on the server """
        if not self.ssh.exists(self.getLibraryIndexFilename()):
            self.dprint("missing library index file [%s|%s]" % (self.getLibraryIndexFilename(),
                self.ssh.normpath(self.getLibraryIndexFilename() )))
            return False
        return True

    def checkHistoryFile(self):
        """ check if history file exists on the server """
        if not self.ssh.exists(self.getHistoryFilename()):
            self.dprint("missing services config file [%s|%s]" % (self.getHistoryFilename(),
                self.ssh.normpath(self.getHistoryFilename() )))
            return False
        return True

    def cleanTemp(self):
        """ clean up all files left in the temp dir
            it should be used before closing a session
        """
        self.dprint("deleting dir [%s]" %  self.getTempPath() )
        self.ssh.delfiles([self.getTempPath()])
        self.dprint("re-creating dir [%s]" %  self.getTempPath() )
        self.ssh.makedir(self.getTempPath())


    def checkServerStatus(self):
        """
        update the self.properties dictionary

        connect to the server and check if the raccoon dir is present
        if present, read the config file:
          locate the lig library path (set the variable if present)
        """
        # check if the server is reachable
        self.isServerAccessible()
        if not self.properties['online']:
            self.dprint("server is not online")
            return False
        # gather system info
        self.getSystemInfo()

        # the status of the server is set to not ready...
        self.properties['ready'] = False
        
        data_to_check = [ [self.checkRaccoonDir, 'Raccoon dir'],
                          [self.checkConfigPath, 'Config dir'],
                          [self.checkLibraryPath, 'Library dir'],
                          [self.checkDataPath, 'Data dir'],
                          [self.checkTempPath, 'Temp dir'],
                          [self.checkBinPath, 'Bin dir'],
                          [self.checkLibraryIndex, 'Library index'],
                          [self.checkHistoryFile, 'History file'],
                        ]

        for checker, msg in data_to_check:
            if not checker():
                self.dprint('%s ***NOT PASSED***' % msg)
                return False
            else:
                self.dprint('%s [ PASSED ]' % msg)
        # ALL TEST PASSED, update status
        # XXX is the scheduler required? XXX
        """
        self.dprint("checking scheduler")
        if not self.properties['scheduler'] == '?':
            # ALL TEST PASSED, update status
            self.properties['ready'] = True
            return True
        else:
            self.dprint("unknown/missing scheduler")
            return False
        """
        self.properties['ready'] = True
        return True
        

    def racconize(self, force=False):
        """ racconize a server if necessary"""

        if self.properties['ready']:
            if not force:
                self.dprint("self.properties report the server as 'ready'")
                print "The server is already racconized. Use 'force=1' to overwrite previous installation"
                return True
        # checking directories
        if not self._checksetupdirs(force=force):
            print "ERROR: Racconization not successful! [step: checking directories]"
            return False
        # check files
        if not self._checksetupfiles(force=force):
            print "ERROR: Racconization not successful! [step: checking files]"
            return False
        print "[ DONE ]"
        self.properties['ready'] = True
        return True



    def _checksetupdirs(self, force=False):
        """ initialize directories that are necessary for the 
            raccoonification process
            
            when force=1 is used, any pre-existing file will be 
            overwritten
        """
                         # checker,              name_getter      comment
        dirs_to_check = [ [self.checkRaccoonDir, self.getRacPath, 'Raccoon directory'],
                          [self.checkConfigPath, self.getConfigPath, 'Config directory'],
                          [self.checkLibraryPath, self.getLibraryPath, 'Library directory'],
                          [self.checkDataPath, self.getDataPath, 'Data directory'],
                          [self.checkTempPath, self.getTempPath, 'Temp directory'],
                          [self.checkBinPath, self.getBinPath, 'Bin directory'],
                        ]
        for checker, dirname, msg in dirs_to_check:
            makedir = True
            path = dirname()
            if checker():
                if force:
                    self.dprint("Directory [%s] already exists [FORCING DELETION...]" %  path )
                    # XXX here a backup could occur...?
                    self.ssh.delfiles([path] )
                    self.dprint("Directory [%s] already exists [ >>DELETED<< ]" %  path )
                else:
                    self.dprint("Directory [%s] already exists [SKIPPING] (use force to overwrite it)" %  path )
                    makedir = False
            if makedir:
                try:
                    self.dprint("Creating directory [%s]" % path)
                    self.ssh.makedir( path )
                except:
                    print "_checksetupdirs> Error creating %s [%s]: %s" % (msg, path, sys.exc_info()[1])
                    return False
        return True


    def _checksetupfiles(self, force=False):
        """ initialize config files that are necessary for the
            raccoonification process
            
            when force=1 is used, any pre-existing file will be 
            overwritten
        """           
                         # checker, name_getter, maker, comment
        files_to_check = [ 
                          ### [ self.checkServicesConfig, self.getServicesConfFilename, 
                          ###  self.makeServicesConf, 'Services configuration file'],
                          ### DISABLED NOW IT IS MANAGED BY THE SERVICE ITSELF

                           [ self.checkLibraryIndex, self.getLibraryIndexFilename, 
                             self.makeLibraryIndex, 'Library index file'],

                           [ self.checkHistoryFile, self.getHistoryFilename, 
                             self.makeHistoryFile, 'History log file'],
                         ]
        for checker, fnamegetter, maker, msg in files_to_check:
            makefile = True
            filename = fnamegetter()
            self.dprint("Checking file [%s]" % filename),
            if checker():
                self.dprint("FOUND"),
                if force:
                    self.dprint("File [%s] already exists [FORCING DELETION]" %  filename )
                    self.ssh.delfiles( [filename] )
                else:
                    self.dprint("File [%s] already exists [SKIPPING] (use force to overwrite it)" %  filename )
                    makefile = True
            if makefile:
                try:
                    self.dprint("Creating file [%s]..." % filename),
                    maker( {'fname': filename })
                    print "[SUCCESS]"
                except:
                    print "_checksetupfiles> Error creating %s [%s]: %s" % (msg, filename, sys.exc_info()[1])
                    return False
        return True


    def makeLibraryIndex(self, library_info={}):
        """ initialize the library index file
            generating an empty file
        """
        fname = library_info['fname']
        self.writetext(fname, '{}')
        #self.ssh.touchfile(fname)

    def makeHistoryFile(self, history_info = {}):
        """ initialize the history log file
            generating an empty file
        """
        fname = history_info['fname']
        self.ssh.touchfile(fname)

    def findLibraries(self):
        """
            scan the master library file for registered 
            libraries on the server
        """
        self.dprint("searching for libraries on the server")
        self.loadLibraryIndex()

    def findServices(self):
        """ scan the directory containing services $RACDIR/$CONFIG
            and return the services found
        """
        if not self.properties['online']:
            print "Not connected to the server yet. Connect to it with .openconnection()"
            return
            
        if not self.properties['ready']:
            print "Server not racconized yet. Prepare it first with .racconize()"
            return

        # delete previous services
        del self.properties['services']
        self.properties['services'] = {}
        del self._unknown_services[:]

        path = self.getConfigPath() 
        config = self.ssh.listdir(path)
        if not config:
            self.dprint("No config files found in %s" % path)
            return False
        self.dprint("FOUND %d FILES %s" % (len(config), config))
        services = []
        problematic = []
        for c in config:
            if c[-5:] == ".conf":
                fname = path + '/' + c
                self.dprint("parsing file [%s]" % fname)
                srv, issues = self.parseServiceConfig(fname)
                if len(issues):
                    self.dprint('problematic lines in file: [%s]' % str(issues))
                if 'service' in srv.keys():
                    if len(srv) > 1: # the kw 'file' is always present
                        services.append(srv)
                        self.dprint('config data accepted')
                    else:
                        problematic.append([fname, 'not enough keywords'])
                        self.dprint('not enough keywords')
                else:
                    problematic.append([fname, 'missing "service" kw'])
                    self.dprint('missing server kw in config file')
        for service_dict in services:
            # register services
            stype = service_dict['service']
            if stype in RaccoonServices.knownservices.keys():
                self.addService(service_dict)
            else:
                print "\n\n==================================================="
                print "WARNING: unrecognized service type: %s" % stype
                print "config file: %s" % service_dict['file']
                print "===================================================\n"
                self.dprint("unknown service[%s] file[%s]" % (stype, service_dict['file']))
                problematic.append([fname, 'unknown service type'])
                self._unknown_services.append( [ stype, service_dict['file'] ] )
        return problematic
                
    def addServiceFromConf(self, fname):
        """ add a service by parsing a config file
        """
        self.dprint("adding service from [%s]" % fname)
        conf, problems = self.parseServiceConfig(self.getConfigPath() + fname)
        print "SERVCONFIG", conf, problems
        return self.addService(service_dict = conf)

    def addService(self, service_dict = {}, allowduplicates=False, debug=False):
        """ add a service to the list of known services 
        
            services are grouped by type
        """
        stype = service_dict['service']

        self.dprint("adding service [%s], type[%s], from config[%s]" % (service_dict['name'],
            stype, service_dict['file']), new=True )
        if not stype in self.properties['services'].keys():
            # first time this server type is found
            self.properties['services'][stype] = []
        status = { 'accepted' : False, 'duplicate' : [] }
        # find possible duplicated services
        all_services_config = [ x.config  for x in self.services(_type= stype) ]
        skip_kw = set([ 'name', 'comment', 'validated', 'file'])

        for conf in all_services_config:
            conf_kw = set(conf.keys())
            new_conf_kw = set(service_dict.keys())
            kw_to_check = (conf_kw - skip_kw) & ( new_conf_kw - skip_kw)
            self.dprint("%d common settings: %s" % (len(kw_to_check), kw_to_check) )
            c = 0
            for kw in kw_to_check:
                if service_dict[kw] == conf[kw]:
                    c+=1
            if len(kw_to_check) == c:
                status['duplicate'].append(conf['file'])
                self.dprint("[file:%s] found a service with identical config: [%s]" % (service_dict['file'], 
                    status['duplicate'][-1]) )
                if allowduplicates: # accept 
                    self.dprint("duplicate found but allowed. Continuing...")
                else: # reject
                    status['accepted'] = False
                    print "duplicate service found, rejected"
                    self.dprint("duplicates not allowed (use allowduplicates=1 to override)")
                    return status
        self.dprint("Registering service...")    
        status['accepted'] = True
        srv = RaccoonServices.knownservices[stype](server=self, config = service_dict, debug=self.debug )
        self.properties['services'][stype].append(srv)
        self.dprint("%d duplicates found: %s" % (len(status['duplicate']), status['duplicate']))
        return status

    def delService(self, service_name, removeconf=True):
        """remove a registered service"""
        self.dprint("asked service_name[%s]" % service_name)
        srv = self.services(name=service_name)
        if len(srv):
            srv = srv[0]
            # remove config file
            if removeconf:
                conf = srv.config['file']
                self.dprint("removing config file [%s]" % conf)
                self.ssh.delfiles([conf])
            # unregister the service from the server
            _type = srv._srvtype
            idx = self.properties['services'][_type].index(srv)
            self.properties['services'][_type].pop(idx)
            self.dprint("service [%s] removed" % srv )
        # EVENT? 


    # XXX NOT USED?
    def writeconfig(self, fname, config):
        """ write service config data in required file
            in the services config location
        """
        # XXX NOT SURE IF THIS SHUOLD BE MANAGED BY 
        #    THE SERVICE ITSELF OR THE SERVER
        fullpath = self.getRacDir() + '/' + fname
        self.dprint('write config file [%s]' % fullpath)
        return self.writetext(fullpath, config)
    # XXX NOT USED

        
    def services(self, _type=None, name=None):
        """"Return all list(services) that have been found and registered

            if _type is requested, return all services matching the request

            if no _type is found or no services are registered, [] is returned.
        """
        if not _type == None and not name == None:
            print "specify either _type or name"
            return False
        if _type:
            return self.properties['services'].get(_type, [])
        srv = []
        for t in self.properties['services']:
            for s in self.properties['services'][t]:
                if (not name == None):
                    if s.config['name'] == name:
                        srv.append(s)
                else:
                    srv.append(s)
        self.dprint("Services %d retrieved" % len(srv))
        return srv

    def service(self, name):
        """ retrieve service by name """
        for t in self.serviceTypes():
            for s in self.properties['services'][t]:
                if s.config['name'] == name:
                    return s
        return None


    def serviceTypes(self):
        """ return found service types """
        return sorted(self.properties['services'].keys())

    def removeService(self, name=None, _type = None):
        """ remove a service for the list of registered services """
        if name == None and _type == None:
            self.dprint("name and _type are None, returning")
            return
        if not name == None and not _type == None:
            self.dprint("both name[%s] and type[%s] requested: conflict!" % (name, _type) )
        # remove by type
        if not _type == None:
            if not _type in self.properties['services'].keys():
                self.dprint( "No services of type [%s]" % _type)
                return
            del self.properties['services'][_type]
            return
        # remove by name
        for t in self.properties['services']:
            for s in self.properties[t]:
                srv = self.properties['services'][t][s]
                if srv.name == name:
                    self.properties['services'][t].pop(srv)
                    return

    def parseServiceConfig(self, fname):
        """ parse service config file and return a 
            dictionary description of the service

            config file has the format:

            kw = value

            it can contains comments: any text after the '#' 
            character will be ignored
        """
        from string import strip
        service = {'file' : fname}
        config = self.ssh.remotereadl(fname, dostrip=1, noempty=1, removecomments=1)
        problems = []

        if config == False:
            print "parseServiceConfig> error parsing config file [%s]" % fname, config
            return False
        for line in config:
            if '=' in line:
                l = line.split("=")
                if len(l) > 2:
                    err =  '[%s] too many "=" characters in line' % line
                    self.dprint(err)
                    issue.append(err)
                    continue
                l = map(strip, l)
                quotes = [ "\"", "\'"]
                # leading quotes
                l = [ x[1:] if x[0] in quotes else x for x in l]
                # trailing quotes
                l = [ x[:-1] if x[-1] in quotes else x for x in l]
                kw, val = map(strip, l)
                # handle multiple values definition for a keyword
                # i.e. library service (each 'lib' kw defines a # library)
                if not kw in service.keys():
                    service[kw] = val
                else:
                    if not isinstance(service[kw], list):
                        service[kw] = [ service[kw] ]
                    # manage multiline comments
                    #if "\\n" in val: val = val.replace('\\n', '\n')
                    service[kw].append(val)
            else:
                err = '[%s] wrong format ("kw = value")' % line
                self.dprint(err)
                problems.append(err)
        return service, problems

    def installBinaryDISABLED(self, url=None, program='vina', tempdir=None, force=False):
        """
        if a custom url is provided, the program type must be specified as well
        """
        data_source = { 'vina' : {'url': 'http://vina.scripps.edu/download/autodock_vina_1_1_2_linux_x86.tgzxxx',
                                'ver': '1.1.2',
                                'binary': ['vina'],
                                'text' : 'vina --version',
                                },
                    'autodock' : {'url': 'http://autodock.scripps.edu/downloads/autodock-registration/tars/dist423/autodocksuite-4.2.3-i86Linux2.tar.gz',
                                  'ver': '4.2.1',
                                  'binary' : ['autodock4', 'autogrid4'],
                                  'test' : 'autodock4 -v',
                                }
                    }

        status = { 'success': False, 'reason' : None }

        if url == None and program == None:
            self.dprint("Null request! link and program == None")
            return False

        info = data_source[program]
        # check if the file is already there
        for bin_name in info['binary']:
            self.dprint('Checking if binary [%s] already installed' % bin_name)
            if self.ssh.exists(bin_name):
                self.dprint('Found! aborting installation')
                return False

        # guessing the program link
        if not url:
            url = info['url']
            self.dprint("downloading %s [%s]" % (program, url))
        else:
            self.dprint("downloading from link: [%s]" % (url))

        # temp working dir and donwnload filename
        if tempdir == None:
            tempdir = os.getcwd()
        ## fname = os.path.basename(url) ! FRAGILE
        fname = url.rsplit('/', 1)[1]
        outfile = tempdir + os.path.sep + fname
        # downloading
        try:
            self.dprint('downloading file as [%s]...' % outfile)
            result = urllib.urlretrieve(url,outfile) # (fname, headers)
            installer, headers = result
        except:
            self.dprint('Error [%s]' % sys.exc_info()[1] )
            return False
        # extracting tar file
        result, source_files = hf.tarextract(tarfilename=installer, 
                                    filelist=info['binary'], outdir= tempdir)
        if result:
            dest_files = [ self.getBinPath() + os.path.basename(f) for f in source_files ]
            files = zip(source_files, dest_files)
            problems = self.ssh.putfiles( flist = files, mode=700) # <-executable
            if len(problems):
                print "PROBLEMS", problems
                self.dprint('Found problems:\n%s' % '\n'.join(problems) )
                return False
        return True

    def getQueueStatus(self):
        """
        this function should collect information on the 
        running jobs, the status of the resources (free nodes,
        etc..)

        it should spawn separate functions depending on the
        scheduler/res manager found
        """
        pass


    def registerJob(self, submission):
        """  register a job in self.history 
            and update the server history file
        """
        # project
        prj = submission.keys()[0]
        if not prj in self.history.keys():
            self.history[prj] = {}
        # experiment
        exp = submission[prj].keys()[0]
        if not exp in self.history[prj].keys():
            self.history[prj][exp] = {} #'vs':{} }
        # jobs
        jobname = submission[prj][exp].keys()[0]
        if jobname in self.history[prj][exp].keys():
            self.dprint("*TROUBLES AHEAD* : name already present [%s]... it should be unregistered first! Now is going to be overwritten." % jobname)
        self.history[prj][exp][jobname] = submission[prj][exp][jobname]
        # save date into remote history file
        self.saveHistory()

    def deleteJobRecord(self, name, prj=None, exp=None):
        """ remove a job record from the history dictionary
            if it is the only item in the experiment and the project,
            remove them too
        """
        data = self.findJob(name)
        if data == None:
            self.dprint("deleting non-existent job... returning")
            return False
        prj, exp, name = data
        self.dprint("deleting job record of [%s]" % name)
        del self.history[prj][exp][jobname]
        if len(self.history[prj][exp].keys()):
            return True
        self.dprint("Exp [%s] is now empty, deleting" % exp)
        del self.history[prj][exp]
        if len(self.history[prj].keys()):
            return True
        self.dprint("Prj [%s] is now empty, deleting" % prj)
        del self.history[prj]
        return True


    def unregisterJob(self, name, prj=None, exp=None, removefiles=True, kill=True):
        """ unregister projects, experiments or single vs jobs
            and update the history file

            by default associated result files, if any, will be deleted
            and running jobs killed
        """
        report = { 'delfiles' : None, 'killjobs' : None}
        if removefiles or kill:
            report['killjobs'] = self.killJob(name)
        if removefiles:
            report['delfiles'] = self.deleteJobData(name)
            #if not report == None:
            #    return report
        self.deleteJobRecord()
        self.saveHistory()
        return True

    def deleteJobData(self, name, prj=None, exp=None):
        """
        """
        data = self.findJob(name)
        data = self.getJobData(data)
        vsdir = data['vsdir']
        self.dprint("deleting vsdir [%s]" % vsdir)
        report = self.ssh.delfiles(vsdir)
        return report

    def killJob(self, name):
        """ try to kill running job """
        # test if job is running?
        # try to kill it
        data = self.findJob(name)
        data = self.getJobData(data)
        jobid = data['jobid']
        vsdir = data['vsdir']
        jobid.reverse() # to remove the collector first
        jobid = " ".join(jobid)
        sched = self.getScheduler()
        if sched == 'pbs':
            cmd = 'qdel %s ; touch %s/.killed'
        elif sched == 'sge':
            cmd = 'qdel %s ; touch %s/.killed'
        cmd = cmd % ( jobid, vsdir)
        report = self.ssh.execmd(cmd)
        return report

    def updateJob(self, name):
        """ update job status """
        data = self.findJob(name)
        if data == ():
            return "unknown job"
        data = self.getJobData(data)
        # check if it has been deleted
        print "DATA", data
        if not self.ssh.exists(data['vsdir']):
            return 'deleted'
        
        # check if it has been killed
        if self.isjobkilled(name):
            return 'killed'

        sched = self.getScheduler()
        if sched == 'pbs' : 
            func = self.pbsQstat
            self.dprint("using PBS scheduler control")
        elif sched == 'sge': 
            func = self.sgeQstat
            self.dprint("using SGE scheduler control")

        # get job status
        total_status = []
        for j in data['jobid']:
            total_status.append(func(j))

        self.dprint("the scheduler reported the following statuses : %s" % total_status)

        # no jobs left
        if total_status == ['completed'] * len(data['jobid']):
            return 'completed'

        # all jobs still queued
        if total_status == ['queued'] * len(data['jobid']):
            return 'queued'

        # one job == collecting
        if len(total_status) == 1:
            coll_status = self.isjobcollecting(name)
            if coll_status == 'collecting':
                return 'collecting'

        # running
        if 'running' in total_status:
            # check completion level
            completion = self.completionLevel(name)
            if completion == 'error':
                return 'error (completion check)' 
            c =  hf.percent(completion, data['total'])
            if c > 100:
                self.dprint("caught count glitch! completion[%2.3f%%]" % c)
                c = 100
            return c
        if 'held' in total_status:
            return 'held'
        return 'dead'


    def isjobkilled(self, name):
        """ return boolean if job has been killed"""
        data = self.findJob(name)
        data = self.getJobData(data)
        # check if the job has been killed
        fname = data['vsdir'] + '/' + '.killed'
        if self.ssh.exists(fname):
            return True
        return False

    def isjobcollecting(self, name):
        """ return boolean if job data is getting collected by tar"""
        data = self.findJob(name)
        data = self.getJobData(data)
        fname = data['vsdir'] + '/' + '.collecting'
        if self.ssh.exist(fname):
            return 'collecting'
        fname = data['vsdir'] + '/' + '.collected'
        if self.ssh.exist(fname):
            return 'done'

    def pbsQstat(self, jobid):
        """ check the status of a job
                "job_state = R"
        """
        kw = 'job_state'
        cmd = """qstat -f  %s | grep %s""" % (jobid, kw)
        report = self.ssh.execmd(cmd)
        if not report: return 'error'
        stdout, stderr = report
        if len(stdout):
            self.dprint("qstat output [%s]" % stdout[0])
            status = stdout[0].split("=",1)[1].strip()
            if status == 'R':
                return 'running'
            elif status == 'Q':
                return 'queued'
            elif status == 'H':
                return 'held'
        else:
            return 'completed'
     

    def sgeQstat(self, jobid):
        """ """
        cmd = """qstat | grep  %s |awk '{print $5}'""" % jobid
        report = self.ssh.execmd(cmd)
        if not report: return 'error'
        stdout, stderr = report
        if len(stdout):
            self.dprint("qstat output [%s]" % stdout[0])
            status = stdout[0].split("=",1)[1].strip()
            status.lower()
            if status == 'r':
                return 'running'
            elif status == 'q':
                return 'queued'
        else:
            return 'completed'



    def completionLevel(self, jobid):
        """ try to calculate completion level"""
        data = self.findJob(jobid)
        data = self.getJobData(data)
        cmd = """wc -l %s/%s""" % (data['vsdir'], data['statusfile'])
        self.dprint("cmd> %s" % cmd)
        report = self.ssh.execmd(cmd)
        self.dprint("report: ", report)
        if not report: 
            return 'error'
        stdout, stderr = report
        if len(stdout):
            count = int(stdout[0].split(" ", 1)[0])
            self.dprint("jobs completed [%d]" % count)
            return count
        return 'error'

            
    def downloadResult(self, name, destpath, bg=True, cb=None):
        """ retrieve the tar file with results""" 
        self.dprint("name[%s] destpath[%s] bg[%s]" % (name, destpath, bg) )
        data = self.findJob(name)
        data = self.getJobData(data)
        tarfile = "%s.tar.gz" % name
        source = "%s/%s" % ( data['vsdir'], tarfile)
        #dest = "%s%s%s" ( destpath, os.sep, tarfile)
        return self.transfer.downloadBigFile(source, destpath, bg, cb) 
    
    def updateDownloadStatus(self):
        """ call the download manager and ask for the status"""
        return self.transfer.updateDownloadBigFile()

    def stopDownload(self):
        """ halt an ongoing download"""
        self.transfer.stop()


    def getMasterLog(self, name):
        """ retrieve the content of the master log file"""
        # scheduler_master.log
        data = self.findJob(name)
        data = self.getJobData(data)
        fname = data['vsdir'] + '/' + data['masterlog']
        opening = self.ssh.openfile(fname)
        if isinstance(opening, tuple):
            fp, sftp = opening
            lines = fp.readlines()
            fp.close()
            sftp.close()
            return lines
        return opening

#############################


class RaccoonLocalServer(DebugObj):

    def __init__(self, debug=True):
        """we need to test the architecture first!"""
        pass 


    def getHomePath(self):
        try:
            from win32com.shell import shellcon, shell            
            homedir = shell.SHGetFolderPath(0, shellcon.CSIDL_APPDATA, 0, 0)
 
        except ImportError: # quick semi-nasty fallback for non-windows/win32com case
            homedir = os.path.expanduser("~")
        
##############################


class RaccoonOpalServer(DebugObj):

    def __init__(self, debug=True):
        """we need to test the architecture first!"""
        pass 




if __name__ == '__main__':
    print "START TESTING"



